#!/usr/bin/env python
# encoding: utf-8
#
# test_spaxels.py
#
# Created by José Sánchez-Gallego on Sep 9, 2016.


from __future__ import division, print_function, absolute_import

import os

import pytest
import astropy.io.fits

from marvin import config
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.modelcube import ModelCube
from marvin.core.exceptions import MarvinError
from marvin.tools.analysis_props import DictOfProperties
from marvin.tools.spaxel import Spaxel
from marvin.tools.spectrum import Spectrum


class TestSpaxelInit(object):

    def _spaxel_init(self, spaxel, cube, maps, spectrum):

        args = {'cube': cube, 'maps': maps, 'spectrum': spectrum}
        objs = {'cube': Cube, 'maps': Maps, 'spectrum': Spectrum}

        for key in objs:
            if args[key]:
                assert isinstance(getattr(spaxel, key), objs[key])
            else:
                assert getattr(spaxel, key) is None

        if maps:
            assert len(spaxel.properties) > 0
            assert isinstance(spaxel.properties, DictOfProperties)
        else:
            assert len(spaxel.properties) == 0

    def test_no_cube_no_maps_db(self, galaxy):
        spaxel = Spaxel(x=15, y=16, plateifu=galaxy.plateifu)
        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_cube_false_no_maps_db(self, galaxy):
        spaxel = Spaxel(x=15, y=16, plateifu=galaxy.plateifu, cube=False)
        self._spaxel_init(spaxel, cube=False, maps=True, spectrum=False)

    def test_no_cube_maps_false_db(self, galaxy):
        spaxel = Spaxel(x=15, y=16, plateifu=galaxy.plateifu, maps=False)
        self._spaxel_init(spaxel, cube=True, maps=False, spectrum=True)

    def test_cube_object_db(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        spaxel = Spaxel(x=15, y=16, cube=cube)

        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_cube_object_maps_false_db(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        spaxel = Spaxel(x=15, y=16, cube=cube, maps=False)

        self._spaxel_init(spaxel, cube=True, maps=False, spectrum=True)

    def test_cube_maps_object_filename(self, galaxy):
        if galaxy.bintype not in ['SPX', 'NONE']:
            pytest.skip("Can't instantiate a Spaxel from a binned Maps.")

        cube = Cube(filename=galaxy.cubepath)
        maps = Maps(filename=galaxy.mapspath, bintype=galaxy.bintype, release=galaxy.release)
        spaxel = Spaxel(x=15, y=16, cube=cube, maps=maps)

        assert cube._drpver == galaxy.drpver
        assert spaxel._drpver == galaxy.drpver
        assert maps._drpver == galaxy.drpver
        assert maps._dapver == galaxy.dapver
        assert spaxel._dapver == galaxy.dapver

        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_cube_object_api(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu, mode='remote')
        spaxel = Spaxel(x=15, y=16, cube=cube)

        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_cube_maps_object_api(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu, mode='remote')
        maps = Maps(plateifu=galaxy.plateifu, mode='remote')
        spaxel = Spaxel(x=15, y=16, cube=cube, maps=maps)

        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_db_maps_template(self, galaxy):
        spaxel = Spaxel(x=15, y=16, cube=False, modelcube=False, maps=True,
                        plateifu=galaxy.plateifu, template_kin=galaxy.template)
        assert spaxel.maps.template_kin == galaxy.template

    def test_api_maps_invalid_template(self, galaxy):
        with pytest.raises(AssertionError) as cm:
            Spaxel(x=15, y=16, cube=False, modelcube=False, maps=True, plateifu=galaxy.plateifu,
                   template_kin='invalid-template')
        assert 'invalid template_kin' in str(cm.value)

    def test_load_false(self, galaxy):
        spaxel = Spaxel(plateifu=galaxy.plateifu, x=15, y=16, load=False)

        assert not spaxel.loaded
        assert spaxel.cube
        assert spaxel.maps
        assert spaxel.modelcube
        assert len(spaxel.properties) == 0
        assert spaxel.spectrum is None

        spaxel.load()

        self._spaxel_init(spaxel, cube=True, maps=True, spectrum=True)

    def test_fails_unbinned_maps(self, galaxy):
        if galaxy.bintype in ['SPX', 'NONE']:
            pytest.skip("Can instantiate a Spaxel from a binned Maps.")

        maps = Maps(plateifu=galaxy.plateifu, bintype=galaxy.bintype, release=galaxy.release)

        with pytest.raises(MarvinError) as cm:
            Spaxel(x=15, y=16, plateifu=galaxy.plateifu, maps=maps)

        assert 'cannot instantiate a Spaxel from a binned Maps.' in str(cm.value)

    def test_spaxel_ra_dec(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        spaxel = Spaxel(x=15, y=16, cube=cube)

        assert pytest.approx(spaxel.ra, 232.54512)
        assert pytest.approx(spaxel.dec, 48.690062)

    @pytest.mark.parametrize('mpl', ['MPL-4', 'MPL-5'])
    def test_release(self, monkeypatch, galaxy, mpl):
        monkeypatch.setattr(config, 'release', mpl)

        cube = Cube(plateifu=galaxy.plateifu)
        spaxel = Spaxel(x=15, y=16, cube=cube)

        assert spaxel.release == mpl

        with pytest.raises(MarvinError) as cm:
            spaxel.release = 'a'
            assert 'the release cannot be changed' in str(cm.value)


class TestPickling(object):

    def test_pickling_db_fails(self, temp_scratch, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        spaxel = cube.getSpaxel(x=1, y=3)

        file = temp_scratch.join('test_spaxel.mpf')

        with pytest.raises(MarvinError) as cm:
            spaxel.save(str(file), overwrite=True)

        assert 'objects with data_origin=\'db\' cannot be saved.' in str(cm.value)

    def test_pickling_only_cube_file(self, temp_scratch, galaxy):
        if galaxy.bintype != 'SPX':
            pytest.skip("Can't instantiate a Spaxel from a binned Maps.")

        cube = Cube(filename=galaxy.cubepath)
        maps = Maps(filename=galaxy.mapspath)

        spaxel = cube.getSpaxel(x=1, y=3, properties=maps, modelcube=False)

        file = temp_scratch.join('test_spaxel.mpf')

        path_saved = spaxel.save(str(file), overwrite=True)
        assert file.check() is True
        assert os.path.exists(path_saved)

        del spaxel

        spaxel_restored = Spaxel.restore(str(file))
        assert spaxel_restored is not None
        assert isinstance(spaxel_restored, Spaxel)

        assert spaxel_restored.cube is not None
        assert spaxel_restored.cube.data_origin == 'file'
        assert isinstance(spaxel_restored.cube.data, astropy.io.fits.HDUList)

        assert spaxel_restored.maps is not None
        assert spaxel_restored.maps.data_origin == 'file'
        assert isinstance(spaxel_restored.maps.data, astropy.io.fits.HDUList)

    @pytest.mark.parametrize('mpl', ['MPL-5'])
    def test_pickling_all_api(self, monkeypatch, temp_scratch, galaxy, mpl):
        monkeypatch.setattr(config, 'release', mpl)
        drpver, __ = config.lookUpVersions()

        cube = Cube(plateifu=galaxy.plateifu, mode='remote')
        maps = Maps(plateifu=galaxy.plateifu, mode='remote')
        modelcube = ModelCube(plateifu=galaxy.plateifu, mode='remote')
        spaxel = cube.getSpaxel(x=1, y=3, properties=maps, modelcube=modelcube)

        assert spaxel.cube.data_origin == 'api'
        assert spaxel.maps.data_origin == 'api'
        assert spaxel.modelcube.data_origin == 'api'

        file = temp_scratch.join('test_spaxel_api.mpf')

        path_saved = spaxel.save(str(file), overwrite=True)
        assert file.check() is True
        assert os.path.exists(path_saved)

        del spaxel

        spaxel_restored = Spaxel.restore(str(file))
        assert spaxel_restored is not None
        assert isinstance(spaxel_restored, Spaxel)

        assert spaxel_restored.cube is not None
        assert isinstance(spaxel_restored.cube, Cube)
        assert spaxel_restored.cube.data_origin == 'api'
        assert spaxel_restored.cube.data is None
        assert spaxel_restored.cube.header['VERSDRP3'] == drpver

        assert spaxel_restored.maps is not None
        assert isinstance(spaxel_restored.maps, Maps)
        assert spaxel_restored.maps.data_origin == 'api'
        assert spaxel_restored.maps.data is None

        assert spaxel_restored.modelcube is not None
        assert isinstance(spaxel_restored.modelcube, ModelCube)
        assert spaxel_restored.modelcube.data_origin == 'api'
        assert spaxel_restored.modelcube.data is None



# Licensed under a 3-clause BSD style license
"""
Marvin is a package intended to simply the access, exploration, and visualization of
the MaNGA dataset for SDSS-IV.  It provides a suite of Python tools, a web interface,
and a REST-like API, under tools/, web/, and api/, respectively.  Core functionality
of Marvin stems from Marvin's Brain.
"""

import os
import re
import warnings
import sys
import marvin
from collections import OrderedDict
from distutils.version import StrictVersion

# Set the Marvin version
try:
    from marvin.version import get_version
except ImportError as e:
    __version__ = 'dev'
else:
    __version__ = get_version()

# Does this so that the implicit module definitions in extern can happen.
from marvin import extern

from marvin.core.exceptions import MarvinUserWarning, MarvinError
from brain.utils.general.general import getDbMachine
from brain import bconfig
from brain.core.core import URLMapDict

# Inits the log
from brain.core.logger import initLog

from astropy.wcs import FITSFixedWarning

# Defines log dir.
if 'MARVIN_LOGS_DIR' in os.environ:
    logFilePath = os.path.join(os.path.realpath(os.environ['MARVIN_LOGS_DIR']), 'marvin.log')
else:
    logFilePath = os.path.realpath(os.path.join(os.environ['HOME'], '.marvin', 'marvin.log'))

log = initLog(logFilePath)

warnings.simplefilter('once')
warnings.filterwarnings('ignore', 'Skipped unsupported reflection of expression-based index')
warnings.filterwarnings('ignore', '(.)+size changed, may indicate binary incompatibility(.)+')
warnings.filterwarnings('ignore', category=FITSFixedWarning)

# Filters for PY3
# TODO: undestand why these warnings are issued and fix the root of the problem (JSG)
warnings.filterwarnings('ignore', 'can\'t resolve package(.)+')
warnings.filterwarnings('ignore', 'unclosed file <_io.TextIOWrapper+')


class MarvinConfig(object):
    ''' Global Marvin Configuration

    The global configuration of Marvin.  Use the config object to globally set options for
    your Marvin session.

    Parameters:
        release (str):
            The release version of the MaNGA data you want to use.  Either MPL or DR.
        download (bool):
            Set to turn on downloading of objects with sdss_access
        use_sentry (bool):
            Set to turn on/off the Sentry error logging.  Default is True.
        add_github_message (bool):
            Set to turn on/off the additional Github Issue message in MarvinErrors. Default is True.
        drpall (str):
            The location to your DRPall file, based on which release you have set.
        mode (str):
            The current mode of Marvin.  Either 'auto', 'remote', or 'local'. Default is 'auto'
        sasurl (str):
            The url of the Marvin API on the Utah Science Archive Server (SAS)
        urlmap (dict):
            A dictionary containing the API routing information used by Marvin
        xyorig (str):
            Globally set the origin point for all your spaxel selections.  Either 'center' or 'lower'.
            Default is 'center'
    '''
    def __init__(self):

        self._drpall = None
        self._inapp = False

        self._urlmap = None
        self._xyorig = None

        self._release = None

        self.vermode = None
        self.download = False
        self.use_sentry = True
        self.add_github_message = True

        self._plantTree()
        self._checkSDSSAccess()
        self._check_manga_dirs()
        self._setDbConfig()
        self._checkConfig()
        self._check_netrc()
        self.setDefaultDrpAll()

    def _checkPaths(self, name):
        ''' Check for the necessary path existence.

            This should only run if someone already has TREE_DIR installed
            but somehow does not have a SAS_BASE_DIR, MANGA_SPECTRO_REDUX, or
            MANGA_SPECTRO_ANALYSIS directory
        '''

        name = name.upper()
        if name not in os.environ:
            if name == 'SAS_BASE_DIR':
                path_dir = os.path.expanduser('~/sas')
            elif name == 'MANGA_SPECTRO_REDUX':
                path_dir = os.path.join(os.path.abspath(os.environ['SAS_BASE_DIR']), 'mangawork/manga/spectro/redux')
            elif name == 'MANGA_SPECTRO_ANALYSIS':
                path_dir = os.path.join(os.path.abspath(os.environ['SAS_BASE_DIR']), 'mangawork/manga/spectro/analysis')

            if not os.path.exists(path_dir):
                warnings.warn('no {0}_DIR found. Creating it in {1}'.format(name, path_dir))
                os.makedirs(path_dir)
            os.environ[name] = path_dir

    def _check_netrc(self):
        """Makes sure there is a valid netrc."""

        netrc_path = os.path.join(os.environ['HOME'], '.netrc')

        if not os.path.exists(netrc_path):
            warnings.warn('cannot find a .netrc file in your HOME directory. '
                          'Remote functionality may not work. Go to '
                          'https://api.sdss.org/doc/manga/marvin/api.html#marvin-authentication '
                          'for more information.', MarvinUserWarning)
            return

        if oct(os.stat(netrc_path).st_mode)[-3:] != '600':
            warnings.warn('your .netrc file has not 600 permissions. Please fix it by '
                          'running chmod 600 ~/.netrc. Authentication will not work with '
                          'permissions different from 600.')

    def _check_manga_dirs(self):
        """Check if $SAS_BASE_DIR and MANGA dirs are defined.
           If they are not, creates and defines them.
        """

        self._checkPaths('SAS_BASE_DIR')
        self._checkPaths('MANGA_SPECTRO_REDUX')
        self._checkPaths('MANGA_SPECTRO_ANALYSIS')

    def setDefaultDrpAll(self, drpver=None):
        """Tries to set the default location of drpall.

        Sets the drpall attribute to the location of your DRPall file, based on the
        drpver.  If drpver not set, it is extracted from the release attribute.  It sets the
        location based on the MANGA_SPECTRO_REDUX environment variable

        Parameters:
            drpver (str):
                The DRP version to set.  Defaults to the version corresponding to config.release.
        """

        if not drpver:
            drpver, __ = self.lookUpVersions(self.release)
        self.drpall = self._getDrpAllPath(drpver)

    def _getDrpAllPath(self, drpver):
        """Returns the default path for drpall, give a certain ``drpver``."""

        if 'MANGA_SPECTRO_REDUX' in os.environ and drpver:
            return os.path.join(os.environ['MANGA_SPECTRO_REDUX'], str(drpver),
                                'drpall-{0}.fits'.format(drpver))
        else:
            raise MarvinError('Must have the MANGA_SPECTRO_REDUX environment variable set')

############ Brain Config overrides ############
# These are configuration parameter defined in Brain.bconfig. We need
# to be able to modify them during run time, so we define properties and
# setters to do that from Marvin.config.

    @property
    def mode(self):
        return bconfig.mode

    @mode.setter
    def mode(self, value):
        bconfig.mode = value

    @property
    def sasurl(self):
        return bconfig.sasurl

    @sasurl.setter
    def sasurl(self, value):
        bconfig.sasurl = value

    @property
    def release(self):
        return self._release

    @release.setter
    def release(self, value):
        value = value.upper()
        if value not in self._mpldict:
            raise MarvinError('trying to set an invalid release version. Valid releases are: {0}'
                              .format(', '.join(sorted(list(self._mpldict)))))
        self._release = value

        drpver = self._mpldict[self.release][0]
        self.drpall = self._getDrpAllPath(drpver)

    @property
    def session_id(self):
        return bconfig.session_id

    @session_id.setter
    def session_id(self, value):
        bconfig.session_id = value

    @property
    def _traceback(self):
        return bconfig.traceback

    @_traceback.setter
    def _traceback(self, value):
        bconfig.traceback = value

#################################################

    @property
    def urlmap(self):
        """Retrieves the URLMap the first time it is needed."""

        if self._urlmap is None or (isinstance(self._urlmap, dict) and len(self._urlmap) == 0):
            try:
                response = Interaction('api/general/getroutemap', request_type='get')
            except Exception as e:
                warnings.warn('Cannot retrieve URLMap. Remote functionality will not work: {0}'.format(e),
                              MarvinUserWarning)
                self._urlmap = URLMapDict()
            else:
                self._urlmap = response.getRouteMap()

        return self._urlmap

    @urlmap.setter
    def urlmap(self, value):
        """Manually sets the URLMap."""
        self._urlmap = value

    @property
    def xyorig(self):
        if not self._xyorig:
            self._xyorig = 'center'

        return self._xyorig

    @xyorig.setter
    def xyorig(self, value):

        assert value.lower() in ['center', 'lower'], 'xyorig must be center or lower.'

        self._xyorig = value.lower()

    @property
    def drpall(self):
        return self._drpall

    @drpall.setter
    def drpall(self, value):
        if os.path.exists(value):
            self._drpall = value
        else:
            self._drpall = None
            warnings.warn('path {0} cannot be found. Setting drpall to None.'
                          .format(value), MarvinUserWarning)

    def _setDbConfig(self):
        ''' Set the db configuration '''
        self.db = getDbMachine()

    def _checkConfig(self):
        ''' Check the config '''
        # set and sort the base MPL dictionary
        mpldict = {'MPL-5': ('v2_0_1', '2.0.2'),
                   'MPL-4': ('v1_5_1', '1.1.1'),
                   'MPL-3': ('v1_3_3', 'v1_0_0'),
                   'MPL-2': ('v1_2_0', None),
                   'MPL-1': ('v1_0_0', None)}  # , 'DR13': ('v1_5_4', None), 'DR14': ('v2_1_1', None)}
        mplsorted = sorted(mpldict.items(), key=lambda p: p[1][0], reverse=True)
        self._mpldict = OrderedDict(mplsorted)

        # Check the versioning config
        if not self.release:
            topkey = list(self._mpldict)[0]
            log.info('No release version set. Setting default to {0}'.format(topkey))
            self.release = topkey

    def setRelease(self, version):
        """Set the release version.

        Parameters:
            version (str):
                The MPL/DR version to set, in form of MPL-X or DRXX.

        Example:
            >>> config.setRelease('MPL-4')
            >>> config.setRelease('DR13')

        """

        version = version.upper()
        self.release = version

    def setMPL(self, mplver):
        """As :func:`setRelease` but check that the version is and MPL."""

        mm = re.search('MPL-([0-9])', mplver)
        assert mm is not None, 'MPL version must be of form "MPL-[X]"'

        if mm:
            self.setRelease(mplver)

    def setDR(self, drver):
        """As :func:`setRelease` but check that the version is and MPL."""

        mm = re.search('DR1([3-9])', drver)
        assert mm is not None, 'DR version must be of form "DR[XX]"'

        if mm:
            self.setRelease(drver)

    def lookUpVersions(self, release=None):
        """Retrieve the DRP and DAP versions that make up a release version.

        Parameters:
            release (str or None):
                The release version. If ``None``, uses the currently set
                ``release`` value.

        Returns:
            drpver, dapver (tuple):
                A tuple of strings of the DRP and DAP versions according
                to the input MPL version

        """

        release = release or self.release

        try:
            drpver, dapver = self._mpldict[release]
        except KeyError:
            raise MarvinError('MPL/DR version {0} not found in lookup table. '
                              'No associated DRP/DAP versions. '
                              'Should they be added?  Check for typos.'.format(release))

        return drpver, dapver

    def lookUpRelease(self, drpver):
        """Retrieve the release version for a given DRP version

        Parameters:
            drpver (str):
                The DRP version to use
        Returns:
            release (str):
                The release version according to the input DRP version
        """

        # Flip the mpldict
        verdict = {val[0]: key for key, val in self._mpldict.items()}

        try:
            release = verdict[drpver]
        except KeyError:
            raise MarvinError('DRP version {0} not found in lookup table. '
                              'No associated MPL version. Should one be added?  '
                              'Check for typos.'.format(drpver))

        return release

    def switchSasUrl(self, sasmode='utah', ngrokid=None, port=5000, test=False):
        ''' Switches the SAS url config attribute

        Easily switch the sasurl configuration variable between
        utah and local.  utah sets it to the real API.  Local switches to
        use localhost.

        Parameters:
            sasmode ({'utah', 'local'}):
                the SAS mode to switch to.  Default is Utah
            ngrokid (str):
                The ngrok id to use when using a 'localhost' sas mode.
                This assumes localhost server is being broadcast by ngrok
            port (int):
                The port of your localhost server
            test (bool):
                If ``True``, sets the Utah sasurl to the test production, test/marvin2
        '''
        assert sasmode in ['utah', 'local'], 'SAS mode can only be utah or local'
        if sasmode == 'local':
            if ngrokid:
                self.sasurl = 'http://{0}.ngrok.io/marvin2/'.format(ngrokid)
            else:
                self.sasurl = 'http://localhost:{0}/marvin2/'.format(port)
        elif sasmode == 'utah':
            marvin_base = 'test/marvin2/' if test else 'marvin2/'
            self.sasurl = 'https://api.sdss.org/{0}'.format(marvin_base)

    def forceDbOff(self):
        ''' Force the database to be turned off '''
        config.db = None
        from marvin import marvindb
        marvindb.forceDbOff()

    def forceDbOn(self):
        ''' Force the database to be reconnected '''
        self._setDbConfig()
        from marvin import marvindb
        marvindb.forceDbOn(dbtype=self.db)

    def _addExternal(self, name):
        ''' Adds an external product into the path '''
        assert type(name) == str, 'name must be a string'
        externdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'extern', name)
        extern_envvar = '{0}_DIR'.format(name.upper())
        os.environ[extern_envvar] = externdir
        pypath = os.path.join(externdir, 'python')
        if os.path.isdir(pypath):
            sys.path.append(pypath)
        else:
            warnings.warn('Python path for external product {0} does not exist'.format(name))

    def _plantTree(self):
        ''' Sets up the sdss tree product root '''
        if 'TREE_DIR' not in os.environ:
            # set up tree using marvin's extern package
            self._addExternal('tree')
            try:
                from tree.tree import Tree
            except ImportError:
                self._tree = None
            else:
                self._tree = Tree(key='MANGA')

    def _checkSDSSAccess(self):
        ''' Checks the client sdss_access setup '''
        if 'SDSS_ACCESS_DIR' not in os.environ:
            # set up sdss_access using marvin's extern package
            self._addExternal('sdss_access')
            try:
                from sdss_access.path import Path
            except ImportError:
                Path = None
            else:
                self._sdss_access_isloaded = True


config = MarvinConfig()

# Inits the Database session and ModelClasses
from marvin.db.marvindb import MarvinDB
marvindb = MarvinDB(dbtype=config.db)

# Init MARVIN_DIR
marvindir = os.environ.get('MARVIN_DIR', None)
if not marvindir:
    moduledir = os.path.dirname(os.path.abspath(__file__))
    marvindir = moduledir.rsplit('/', 2)[0]
    os.environ['MARVIN_DIR'] = marvindir

# Inits the URL Route Map
from marvin.api.api import Interaction
config.sasurl = 'https://api.sdss.org/marvin2/'




import marvin
import traceback
import sys
import inspect

# This line makes sure that "from marvin.utils.db.dbutils import *"
# will only import the functions in the list.
__all__ = ['get_traceback', 'testDbConnection', 'generateClassDict']


def get_traceback(asstring=None):
    ''' Returns the traceback from an exception, a list

        Parameters:
        asstring = boolean to return traceback as a joined string
    '''
    ex_type, ex_info, tb = sys.exc_info()
    newtb = traceback.format_tb(tb)
    return ' '.join(newtb) if asstring else newtb


def testDbConnection(session=None):
    ''' Test the DB connection to '''

    res = {'good': None, 'error': None}

    if not session:
        session = marvin.marvindb.session

    try:
        tmp = session.query(marvin.marvindb.datadb.PipelineVersion).first()
        res['good'] = True
    except Exception as e:
        error1 = 'Error connecting to manga database: {0}'.format(str(e))
        tb = get_traceback(asstring=True)
        error2 = 'Full traceback: {0}'.format(tb)
        error = ' '.join([error1, error2])
        res['error'] = error

    return res


def generateClassDict(modelclasses, lower=None):
    ''' Generates a dictionary of the Model Classes, based on class name as key, to the object class.
        Selects only those classes in the module with attribute __tablename__
        lower = True makes class name key all lowercase
    '''

    classdict = {}
    for model in inspect.getmembers(modelclasses, inspect.isclass):
        keyname = model[0].lower() if lower else model[0]
        if hasattr(model[1], '__tablename__'):
            classdict[keyname] = model[1]
    return classdict



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-19 17:47:12
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-29 20:11:29

from __future__ import print_function, division, absolute_import
from marvin.tests.api.conftest import ApiPage
from marvin.tests import marvin_test_if
import pytest


@pytest.mark.parametrize('page', [('api', 'getRSS')], ids=['getrss'], indirect=True)
class TestGetRss(object):

    @marvin_test_if(mark='skip', galaxy={'plateifu': ['7443-12701']})
    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_plateifu_success(self, galaxy, page, params, reqtype):
        data = {}
        page.load_page(reqtype, page.url.format(name=galaxy.plateifu), params=params)
        page.assert_success(data)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                       ('badname', 'name', 'String does not match expected pattern.'),
                                                       ('84', 'name', 'Shorter than minimum length 4.')],
                             ids=['norelease', 'badname', 'shortname'])
    def test_plateifu_failure(self, galaxy, page, reqtype, params, name, missing, errmsg):
        if name is None:
            page.route_no_valid_params(page.url.format(name=galaxy.plateifu), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(name=name), missing, reqtype=reqtype, params=params, errmsg=errmsg)


@pytest.mark.slow
@pytest.mark.parametrize('page', [('api', 'getRSSAllFibers')], ids=['getrssfibers'], indirect=True)
class TestGetRssFibers(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_plateifu_success(self, galaxy, page, params, reqtype):
        data = {'wavelength': []}
        page.load_page(reqtype, page.url.format(name=galaxy.plateifu), params=params)
        page.assert_success()
        assert len(page.json['data'].keys()) == 172

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                       ('badname', 'name', 'String does not match expected pattern.'),
                                                       ('84', 'name', 'Shorter than minimum length 4.')],
                             ids=['norelease', 'badname', 'shortname'])
    def test_plateifu_failure(self, galaxy, page, reqtype, params, name, missing, errmsg):
        if name is None:
            page.route_no_valid_params(page.url.format(name=galaxy.plateifu), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(name=name), missing, reqtype=reqtype, params=params, errmsg=errmsg)



#!/usr/bin/python

''' This file handles a database connection. It can simply be deleted if not needed.

    The example given is for a PostgreSQL database, but can be modified for any other.
'''
from __future__ import print_function
from __future__ import division

from marvin.db.DatabaseConnection import DatabaseConnection
from pgpasslib import getpass
from marvin import config
import yaml
import os

# Read in the db connection configuration
dbconfigfile = 'dbconfig.ini'
dbconfigfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), dbconfigfile)

try:
    rawfile = open(dbconfigfile, 'r').read()
except IOError as e:
    raise RuntimeError('IOError: Could not open dbconfigfile {0}:{1}'.format(dbconfigfile, e))
dbdict = yaml.load(rawfile)

# select the appropriate configuration from config
if config.db:
    db_info = dbdict[config.db]
    try:
        if 'password' not in db_info:
            db_info['password'] = getpass(db_info['host'], db_info['port'], db_info['database'], db_info['user'])
    except KeyError:
        raise RuntimeError('ERROR: invalid server configuration')
else:
    raise RuntimeError('Error: could not determine db to connect to: {0}'.format(config.db))

# this format is only usable with PostgreSQL 9.2+
# dsn = "postgresql://{user}:{password}@{host}:{port}/{database}".format(**db_info)
# database_connection_string = 'postgresql+psycopg2://%s:%s@%s:%s/%s' % (db_info["user"], db_info["password"], db_info["host"], db_info["port"], db_info["database"])

# Build the database connection string
if db_info["host"] == 'localhost':
    database_connection_string = 'postgresql+psycopg2:///%(database)s' % db_info
else:
    database_connection_string = 'postgresql+psycopg2://%(user)s:%(password)s@%(host)s:%(port)i/%(database)s' % db_info

# Make a database connection
try:
    db = DatabaseConnection()
except AssertionError as e:
    db = DatabaseConnection(database_connection_string=database_connection_string)
    engine = db.engine
    metadata = db.metadata
    Session = db.Session
    Base = db.Base
except KeyError as e:
    print("Necessary configuration value not defined.")
    raise RuntimeError('KeyError: Necessary configuration value not defined: {0}'.format(e))



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
#
# $Id$
#
"""Create files that can be used to set environment variables.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
#
# Top-level definitions.
#
__author__ = 'Benjamin Weaver <benjamin.weaver@nyu.edu>'
__version__ = '$Revision$'.split(': ')[1].split()[0]
__all__ = [ ]
__docformat__ = 'restructuredtext en'
#
# Modules
#
import argparse
from collections import OrderedDict
import ConfigParser
import glob
import os
import os.path
import sys
#
# to_file()
#
def to_file(env,header='',setenv='setenv {0} {1}'):
    """Convert config file data to module file.

    Parameters
    ----------
    env : collections.OrderedDict
        A mapping containing all the configuration file data.
    header : str
        String to add to the header of the file produced.
    setenv : str
        The method for setting environment variables in this file type.

    Returns
    -------
    lines : str
        A string representation of the file.
    """
    if len(header) == 0:
        lines = """#
# This file sets up tree/{0}.
#
""".format(env['default']['name'])
    else: lines = header.format(**env['default'])
    for section in env:
        if section == 'default': continue
        lines += """#
# {0}
#
""".format(section)
        setenvs = list()
        for k in env[section]: setenvs.append(setenv.format(k,env[section][k]))
        lines += "\n".join(setenvs)
        lines += "\n"
    return lines
#
# parse_cfg()
#
def parse_cfg(cfg,root):
    """Parse a tree configuration file.

    Parameters
    ----------
    cfg : ConfigParser.ConfigParser
        A ``ConfigParser`` object.
    root : str
        The value of ``$SAS_BASE_DIR``.

    Returns
    -------
    env : collections.OrderedDict
        A mapping containing all the configuration file data
    """
    env = OrderedDict()
    replace = '@FILESYSTEM@'
    env['default'] = cfg.defaults()
    if env['default']['FILESYSTEM'] == replace: env['default']['FILESYSTEM'] = root
    env['default']['current'] = env['default']['current'] == 'True'
    for sec in cfg.sections():
        env[sec] = OrderedDict()
        for opt in cfg.options(sec):
            if opt in env['default']: continue
            val = cfg.get(sec,opt)
            print("section=%r opt=%r val=%r" % (sec,opt,val))
            if val.find(replace) == 0: val = val.replace(replace,root)
            env[sec][opt] = val
    return env
#
# Main function
#
def main():
    """Program to run if called as an executable.
    """
    #
    # Get options
    #
    parser = argparse.ArgumentParser(description=__doc__,prog=os.path.basename(sys.argv[0]))
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
        help='Print extra information.')
    parser.add_argument('-r', '--root', action='store', dest='root',
        default=os.getenv('SAS_BASE_DIR'),
        help='Override the value of $SAS_BASE_DIR.',metavar='SAS_BASE_DIR')
    parser.add_argument('-t', '--treedir', action='store', dest='treedir',
        default=os.getenv('TREE_DIR'),
        help='Override the value of $TREE_DIR.',metavar='TREE_DIR')
    options = parser.parse_args()

    #
    # Configure output files
    #
    tcshheader = """# Set up tree/{name} for (t)csh.
setenv TREE_DIR """ + options.treedir.rstrip('/') + """
setenv TREE_VER {name}
setenv PATH $TREE_DIR/bin:$PATH
"""
    bashheader = """# Set up tree/{name} for (ba)sh.
export TREE_DIR=""" + options.treedir.rstrip('/') + """
export TREE_VER={name}
export PATH=$TREE_DIR/bin:$PATH
"""
    moduleheader = """#%Module1.0
proc ModulesHelp {{ }} {{
    global product version
    puts stderr "This module adds $product $version to various paths"
}}
set product tree
set version {name}
conflict $product
module-whatis   "Sets up $product $version in your environment"

set PRODUCT_DIR """ + options.treedir.rstrip('/') + """
setenv [string toupper $product]_DIR $PRODUCT_DIR
setenv [string toupper $product]_VER $version
prepend-path PATH $PRODUCT_DIR/bin
"""
    modulesversion = """#%Module1.0
set ModulesVersion {name}
"""
    outputs = {
        'tcsh':{
            'ext':'.csh',
            'header':tcshheader,
            'setenv':'setenv {0} {1}'},
        'bash':{
            'ext':'.sh',
            'header':bashheader,
            'setenv':'export {0}={1}'},
        'module':{
            'ext':'.module',
            'header':moduleheader,
            'setenv':'setenv {0} {1}'},
        }
    #
    # Find the data directory
    #
    datadir = os.path.join(options.treedir,'data')
    etcdir = os.path.join(options.treedir,'etc')
    if not os.path.exists(datadir):
        datadir = os.path.join('..','data')
        etcdir = '.'
        if not os.path.exists(datadir):
            print("Could not find a data directory!")
            return 1
    #
    # Read the configuration files
    #
    for cfgfile in glob.glob(os.path.join(datadir,'*.cfg')):
        cfg = ConfigParser.SafeConfigParser()
        cfg.optionxform = str
        cfg.read(cfgfile)
        env = parse_cfg(cfg,options.root)
        if options.verbose:
            print(env)
        for output in outputs:
            filedata = to_file(env,outputs[output]['header'],outputs[output]['setenv'])
            filename = os.path.join(etcdir, env['default']['name']+outputs[output]['ext'])
            with open(filename,'w') as f: f.write(filedata)
            if output == 'module' and env['default']['current']:
                versionname = os.path.join(etcdir, '.version')
                with open(versionname,'w') as f: f.write(modulesversion.format(**env['default']))
    return 0
#
#
#
if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# encoding: utf-8
#
# test_maps.py
#
# @Author: Brett Andrews <andrews>
# @Date:   2017-05-01 09:07:00
# @Last modified by:   andrews
# @Last modified time: 2017-07-19 10:07:33

import numpy as np
import matplotlib
import pytest

from marvin import config
from marvin.tools.maps import Maps
import marvin.utils.plot.map as mapplot
from marvin.utils.general import get_plot_params


matplotlib_2 = pytest.mark.skipif(int(matplotlib.__version__.split('.')[0]) <= 1,
                                  reason='matplotlib-2.0 or higher required')

value = np.array([[1, -2, 3],
                  [-4, 5, 6],
                  [7, 8, 9]])

ivar = np.array([[4, 10, 1],
                 [0.01, 0.04, 0.0001],
                 [100000, 0, 1]])

ivar_mpl4 = np.array([[0, 0, 0],
                      [0, 3, 0],
                      [0, 0, 0.5]])

mask_simple = np.array([[0, 1, 0],
                        [1, 0, 1],
                        [0, 1, 0]], dtype=bool)

mask_binary = np.array([[0b000, 0b001, 0b010],
                        [0b011, 0b100, 0b101],
                        [0b110, 0b111, 0b000]])

mask_daplike = np.array([[0b0000000000000000000000000000000,
                          0b0000000000000000000000000000001,
                          0b0000000000000000000000000100000],
                         [0b1000000000000000000000001000001,
                          0b1000000000000000000000010000000,
                          0b1000000000000000000000000000001],
                         [0b0000000000000000000000000000000,
                          0b1000000000000000000000011100001,
                          0b0000000000000000000000000000000]])

nocov = np.array([[0, 1, 0],
                  [1, 0, 1],
                  [0, 1, 0]], dtype=bool)

nocov_mpl4 = np.array([[1, 1, 1],
                       [1, 0, 1],
                       [1, 1, 0]], dtype=bool)

bad_data = np.array([[0, 0, 1],
                     [1, 1, 1],
                     [0, 1, 0]])

bad_data_mpl4 = np.array([[0, 1, 0],
                          [1, 0, 1],
                          [0, 1, 0]], dtype=bool)

# SNR = value * np.sqrt(ivar)
# [[  2.00e+00,  -6.32e+00,   3.00e+00],
#  [ -4.00e-01,   1.00e+00,   6.00e-02],
#  [  2.21e+03,   0.00e+00,   9.00e+00]]

snr_min_none = np.array([[0, 0, 0],
                         [0, 0, 0],
                         [0, 1, 0]])

snr_min_0 = np.array([[0, 0, 0],
                      [0, 0, 0],
                      [0, 1, 0]])

snr_min_1 = np.array([[0, 0, 0],
                      [1, 0, 1],
                      [0, 1, 0]])

snr_min_3 = np.array([[1, 0, 0],
                      [1, 1, 1],
                      [0, 1, 0]])

log_cb_true = np.array([[0, 1, 0],
                        [1, 0, 0],
                        [0, 0, 0]])

log_cb_false = np.zeros(value.shape, dtype=bool)

image_none_true = np.array([[0, 1, 1],
                            [1, 1, 1],
                            [0, 1, 0]])

image_3_true = np.array([[1, 1, 1],
                         [1, 1, 1],
                         [0, 1, 0]])

image_none_false = np.array([[0, 1, 1],
                             [1, 1, 1],
                             [0, 1, 0]])

image_3_false = np.array([[1, 1, 1],
                          [1, 1, 1],
                          [0, 1, 0]])


@pytest.fixture(scope='module', params=['stellar_vel', 'stellar_sigma', 'emline_gflux',
                                        'specindex'])
def bits(request, set_release):
    params = get_plot_params(dapver=config.lookUpVersions()[1], prop=request.param)
    return params['bitmasks']


class TestMasks(object):

    @pytest.mark.parametrize('mask, ivar, expected',
                             [(mask_simple, ivar_mpl4, nocov_mpl4)])
    def test_no_coverage_mask_mpl4(self, mask, ivar, expected):
        if config.release != 'MPL-4':
            pytest.skip('Use generic test for NOCOV bitmask in MPL-5+.')
        nocov = mapplot.no_coverage_mask(mask=mask, bit=None, ivar=ivar)
        assert np.all(nocov == expected)

    @pytest.mark.parametrize('mask, ivar, expected',
                             [(mask_simple, ivar, nocov),
                              (mask_binary, ivar, nocov),
                              (mask_daplike, ivar, nocov)])
    def test_no_coverage_mask(self, mask, bits, ivar, expected):
        if config.release == 'MPL-4':
            pytest.skip('NOCOV bitmask only exists in MPL-5+')
        nocov = mapplot.no_coverage_mask(mask=mask, bit=bits['nocov'], ivar=ivar)
        assert np.all(nocov == expected)

    @pytest.mark.parametrize('mask, expected', [(mask_simple, bad_data_mpl4)])
    def test_bad_data_mask_mpl4(self, mask, expected):
        if config.release != 'MPL-4':
            pytest.skip('Use generic test for bad data bitmasks in MPL-5+.')
        bad_data = mapplot.bad_data_mask(mask=mask, bits={'doNotUse': 0})
        assert np.all(bad_data == expected)

    @pytest.mark.parametrize('mask, expected', [(mask_daplike, bad_data)])
    def test_bad_data_mask(self, mask, bits, expected):
        if config.release == 'MPL-4':
            pytest.skip('UNRELIABLE bitmask only exists in MPL-5+')
        bad_data = mapplot.bad_data_mask(mask=mask, bits=bits['badData'])
        assert np.all(bad_data == expected)

    @pytest.mark.parametrize('value, ivar, snr_min, expected',
                             [(value, ivar, None, snr_min_none),
                              (value, ivar, 0, snr_min_0),
                              (value, ivar, 1, snr_min_1),
                              (value, ivar, 3, snr_min_3),
                              (value, None, None, np.zeros(value.shape, dtype=bool)),
                              (value, None, 1, np.zeros(value.shape, dtype=bool))])
    def test_low_snr_mask(self, value, ivar, snr_min, expected):
        low_snr = mapplot.low_snr_mask(value, ivar, snr_min)
        assert np.all(low_snr == expected)

    @pytest.mark.parametrize('value, log_cb, expected',
                             [(value, True, log_cb_true),
                              (value, False, log_cb_false)])
    def test_log_colorbar_mask(self, value, log_cb, expected):
        log_colorbar = mapplot.log_colorbar_mask(value, log_cb)
        assert np.all(log_colorbar == expected)

    @pytest.mark.parametrize('value, nocov, bad_data, snr_min, log_cb, expected_mask',
                             [(value, nocov, bad_data, snr_min_none, log_cb_true, image_none_true),
                              (value, nocov, bad_data, snr_min_3, log_cb_true, image_3_true),
                              (value, nocov, bad_data, snr_min_none, log_cb_false,
                               image_none_false),
                              (value, nocov, bad_data, snr_min_3, log_cb_false, image_3_false)])
    def test_select_good_spaxels(self, value, nocov, bad_data, snr_min, log_cb, expected_mask):
        image = mapplot.select_good_spaxels(value, nocov, bad_data, snr_min, log_cb)
        expected = np.ma.array(value, mask=expected_mask)
        assert np.all(image.data == expected.data)
        assert np.all(image.mask == expected.mask)


class TestMapPlot(object):

    @pytest.mark.parametrize('cube_size, sky_coords, expected',
                             [([36, 36], True, np.array([-18, 18, -18, 18])),
                              ([35, 35], True, np.array([-17.5, 17.5, -17.5, 17.5])),
                              ([36, 36], False, np.array([0, 35, 0, 35]))])
    def test_set_extent(self, cube_size, sky_coords, expected):
        extent = mapplot.set_extent(cube_size, sky_coords)
        assert np.all(extent == expected)

    @matplotlib_2
    def test_set_hatch_linewidth(self, maps):
        map_ = maps.getMap('emline_gflux', channel='ha_6564')
        fig, ax = mapplot.plot(dapmap=map_)
        assert matplotlib.rcParams['hatch.linewidth'] == 0.5

    @matplotlib_2
    def test_set_hatch_color(self, maps):
        map_ = maps.getMap('emline_gflux', channel='ha_6564')
        fig, ax = mapplot.plot(dapmap=map_)
        assert matplotlib.rcParams['hatch.color'] == 'w'

    @matplotlib_2
    def test_ax_facecolor(self):
        fig, ax = mapplot.ax_setup(sky_coords=True, fig=None, ax=None, facecolor='#A8A8A8')
        assert ax.get_facecolor() == (0.6588235294117647, 0.6588235294117647, 0.6588235294117647,
                                      1.0)

    @pytest.mark.parametrize('sky_coords, expected',
                             [(True, 'arcsec'),
                              (False, 'spaxel')])
    def test_ax_labels(self, sky_coords, expected):
        fig, ax = mapplot.ax_setup(sky_coords, fig=None, ax=None, facecolor='#A8A8A8')
        assert ax.get_xlabel() == expected
        assert ax.get_ylabel() == expected

    @pytest.mark.parametrize('title, property_name, channel, expected',
                             [('test-title', None, None, 'test-title'),
                              ('test-title', 'prop', 'channel', 'test-title'),
                              (None, 'prop', 'channel', 'prop channel'),
                              (None, 'emline_gflux', 'ha_6564', 'emline gflux ha 6564'),
                              (None, 'prop_flux', None, 'prop flux'),
                              (None, None, 'channel', 'channel'),
                              (None, None, None, '')])
    def test_set_title(self, title, property_name, channel, expected):
        title = mapplot.set_title(title, property_name, channel)
        assert title == expected

    @pytest.mark.parametrize('property_name, channel',
                             [('emline_gflux', 'ha_6564'),
                              ('emline_gflux', 'oiii_5008'),
                              ('stellar_vel', None),
                              ('stellar_sigma', None),
                              ('specindex', 'd4000')])
    def test_plot_dapmap(self, maps, property_name, channel):
        map_ = maps.getMap(property_name, channel=channel)
        fig, ax = mapplot.plot(dapmap=map_)
        assert isinstance(fig, matplotlib.figure.Figure)
        assert isinstance(ax, matplotlib.axes._axes.Axes)

    @pytest.mark.parametrize('mask', [(None), (mask_simple), (mask_binary), (mask_daplike)])
    @pytest.mark.parametrize('ivar', [(None), (ivar)])
    @pytest.mark.parametrize('value', [(value)])
    def test_plot_user_defined_map(self, value, ivar, mask):
        fig, ax = mapplot.plot(value=value, ivar=ivar, mask=mask)
        assert isinstance(fig, matplotlib.figure.Figure)
        assert isinstance(ax, matplotlib.axes._axes.Axes)

    def test_plot_matplotlib_rc_restore(self, maps):
        rcparams = matplotlib.rc_params()
        map_ = maps.getMap('emline_gflux', channel='ha_6564')
        fig, ax = mapplot.plot(dapmap=map_)
        assert rcparams == matplotlib.rc_params()

    def test_plot_matplotlib_style_sheet(self, maps):
        map_ = maps.getMap('emline_gflux', channel='ha_6564')
        fig, ax = mapplot.plot(dapmap=map_, plt_style='ggplot')
        assert isinstance(fig, matplotlib.figure.Figure)
        assert isinstance(ax, matplotlib.axes._axes.Axes)

    @pytest.mark.parametrize('title, expected',
                             [('emline_gflux_ha_6564', 'default'),
                              ('emline_gvel_ha_6564', 'vel'),
                              ('emline_gsigma_hb_4862', 'sigma'),
                              ('stellar_vel', 'vel'),
                              ('stellar_sigma', 'sigma')])
    def test_get_prop(self, title, expected):
        assert mapplot._get_prop(title) == expected
    
    def test_return_cb(self, maps):
        map_ = maps.getMap('emline_gflux', channel='ha_6564')
        fig, ax, cb = mapplot.plot(dapmap=map_, return_cb=True)
        assert isinstance(cb, matplotlib.colorbar.Colorbar)




# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-03-24 12:22:30
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-04-04 10:30:44

from __future__ import print_function, division, absolute_import
import os
import pytest

from sdss_access import RsyncAccess

releases = ['MPL-4']
plateifus = ['8485-1901']
mpldict = {'MPL-4': ('v1_5_1', '1.1.1'), 'MPL-5': ('v2_0_1', '2.0.2')}
surveys = ['manga']


@pytest.fixture(scope='session', params=releases)
def get_release(request):
    ''' MaNGA release fixture '''
    return request.param


@pytest.fixture(scope='session', params=plateifus)
def get_plateifu(request):
    ''' MaNGA plateifu fixture '''
    return request.param


class Survey(object):
    ''' Survey class '''
    def __init__(self, survey):
        self.survey = survey
        if self.survey == 'manga':
            self.setup_manga()

    def setup_manga(self):
        self.rsync_kwargs = self.get_rsync_kwargs()

    def get_rsync_kwargs(self):
        if self.survey == 'manga':
            rkwargs = {'plate': None, 'ifu': None, 'drpver': None, 'dapver': None, 'dir3d': None,
                       'mpl': None, 'bintype': '*', 'n': '**', 'mode': '*', 'daptype': '*'}
        return rkwargs

    def set_data(self, release):
        if self.survey == 'manga':
            self.release = release
            self.names = ['mangacube']
            if '4' in release:
                self.names.extend(['mangadefault', 'mangamap'])
            elif '5' in release:
                self.names.extend(['mangadap5'])


@pytest.fixture(scope='session', params=surveys)
def survey(request):
    ''' fixture to generate a survey'''
    survey = Survey(request.param)
    yield survey
    survey = None


@pytest.fixture(scope='session')
def init_manga_survey(survey, get_release, get_plateifu):
    ''' create different surveys with different parameters '''

    drpver, dapver = mpldict[get_release]
    plate, ifu = get_plateifu.split('-')
    survey.rsync_kwargs['drpver'] = drpver
    survey.rsync_kwargs['dapver'] = dapver
    survey.rsync_kwargs['plate'] = plate
    survey.rsync_kwargs['ifu'] = ifu
    survey.set_data(get_release)
    yield survey


@pytest.fixture(scope='module', params=['mangadefault', 'mangamap', 'mangacube'])
def data(request, init_manga_survey):
    ''' fixture to generate data '''
    fillkwargs = {'plate': init_manga_survey.rsync_kwargs['plate'], 'ifu': init_manga_survey.rsync_kwargs['ifu'],
                  'drpver': init_manga_survey.rsync_kwargs['drpver'], 'dapver': init_manga_survey.rsync_kwargs['dapver']}
    data_dict = {'mangadefault': {'files': 'mangadap-{plate}-{ifu}-default.fits.gz'.format(**fillkwargs),
                                  'loc': 'mangawork/manga/spectro/analysis/{drpver}/{dapver}/default/{plate}/'.format(**fillkwargs),
                                  'single': 'mangadap-{plate}-{ifu}-default.fits.gz'.format(**fillkwargs),
                                  'count': 1},
                 'mangamap': {'files': 'manga-{plate}-{ifu}-LOGCUBE_MAPS-*-0**.fits.gz'.format(**fillkwargs),
                              'loc': 'mangawork/manga/spectro/analysis/{drpver}/{dapver}/full/{plate}/{ifu}/'.format(**fillkwargs),
                              'single': 'manga-{plate}-{ifu}-LOGCUBE_MAPS-NONE-003.fits.gz'.format(**fillkwargs),
                              'count': 21},
                 'mangadap5': {'files': 'manga-{plate}-{ifu}-*-SPX-GAU-MILESHC.fits.gz'.format(**fillkwargs),
                               'loc': 'mangawork/manga/spectro/analysis/{drpver}/{dapver}/*/{plate}/{ifu}/'.format(**fillkwargs),
                               'single': 'manga-{plate}-{ifu}-MAPS-SPX-GAU-MILESHC.fits.gz'.format(**fillkwargs),
                               'count': 2},
                 'mangacube': {'files': 'manga-{plate}-{ifu}-LOGCUBE.fits.gz'.format(**fillkwargs),
                               'loc': 'mangawork/manga/spectro/redux/{drpver}/{plate}/stack/'.format(**fillkwargs),
                               'single': 'manga-{plate}-{ifu}-LOGCUBE.fits.gz'.format(**fillkwargs),
                               'count': 1}
                 }
    return (request.param, data_dict[request.param])


@pytest.fixture(scope='function')
def rsync():
    ''' fixture to create generic rsync object '''
    rsync = RsyncAccess(label='test_rsync')
    rsync.remote()
    yield rsync

    # teardown
    rsync.reset()
    rsync = None


@pytest.fixture(scope='function')
def rsync_add(rsync, data, init_manga_survey):
    ''' fixture to add data to an rsync object '''
    name, paths = data
    rsync.add(name, **init_manga_survey.rsync_kwargs)
    rsync.location = os.path.join(paths['loc'], paths['files'])
    yield rsync


@pytest.fixture(scope='function')
def rsync_set(rsync_add, data):
    ''' fixture to set the stream of an rsync object '''
    name, paths = data
    rsync_add.location = os.path.join(paths['loc'], paths['single'])
    rsync_add.count = paths['count']
    rsync_add.set_stream()
    yield rsync_add





#!/usr/bin/python
# encoding: utf-8
'''
Licensed under a 3-clause BSD license.
'''

from __future__ import print_function, division
from flask_restful import Api
from flask import Flask, Blueprint, send_from_directory
from flask import session, request, render_template, g, jsonify
import sys
import os
import logging
import flask_profiler
import flask_jsglue as jsg
# Marvin imports
from brain.utils.general.general import getDbMachine
from marvin import config, log
from marvin.web.web_utils import updateGlobalSession
from marvin.web.jinja_filters import jinjablue
from marvin.web.error_handlers import errors
from marvin.web.extensions import jsglue, flags, sentry, limiter
from marvin.web.settings import ProdConfig, DevConfig, CustomConfig
# Web Views
from marvin.web.controllers.index import index
from marvin.web.controllers.galaxy import galaxy
from marvin.web.controllers.search import search
from marvin.web.controllers.plate import plate
from marvin.web.controllers.images import images
from marvin.web.controllers.users import users
# API Views
from marvin.api.cube import CubeView
from marvin.api.maps import MapsView
from marvin.api.modelcube import ModelCubeView
from marvin.api.plate import PlateView
from marvin.api.rss import RSSView
from marvin.api.spaxel import SpaxelView
from marvin.api.query import QueryView
from marvin.api.general import GeneralRequestsView

# ================================================================================


def create_app(debug=False, local=False, object_config=None):

    # ----------------------------------
    # Create App
    marvin_base = os.environ.get('MARVIN_BASE', 'marvin2')
    app = Flask(__name__, static_url_path='/{0}/static'.format(marvin_base))
    api = Blueprint("api", __name__, url_prefix='/{0}/api'.format(marvin_base))
    app.debug = debug

    # Add Marvin Logger
    app.logger.addHandler(log)

    # Turn on debug stuff in config
    if app.debug is True:
        sentry = None
        config.use_sentry = False
        config.add_github_message = False

    # Find which connection to make
    connection = getDbMachine()
    local = (connection == 'local') or local

    # ----------------------------------
    # Set some other variables
    config._inapp = True
    url_prefix = '/marvin2' if local else '/{0}'.format(marvin_base)

    # ----------------------------------
    # Load the appropriate Flask configuration object for debug or production
    if not object_config:
        if app.debug or local:
            app.logger.info('Loading Development Config!')
            object_config = type('Config', (DevConfig, CustomConfig), dict())
        else:
            app.logger.info('Loading Production Config!')
            object_config = type('Config', (ProdConfig, CustomConfig), dict())
    app.config.from_object(object_config)

    # ------------------------------------------
    # Add lib directory as a new static path
    @app.route('/{0}/lib/<path:filename>'.format(marvin_base))
    def lib(filename):
        return send_from_directory(app.config["LIB_PATH"], filename)

    # Register global session update before every request
    @app.before_request
    def global_update():
        ''' updates the global session / config '''
        updateGlobalSession()

    # ----------------------------------
    # Registration
    register_extensions(app, app_base=marvin_base)
    register_api(app, api)
    register_blueprints(app, url_prefix=url_prefix)

    return app


# ================================================================================

def register_api(app, api):
    ''' Register the Flask API routes used '''

    CubeView.register(api)
    MapsView.register(api)
    ModelCubeView.register(api)
    PlateView.register(api)
    RSSView.register(api)
    SpaxelView.register(api)
    GeneralRequestsView.register(api)
    QueryView.register(api)
    limiter.limit("200/minute")(api)
    app.register_blueprint(api)


def register_extensions(app, app_base=None):
    ''' Register the Flask extensions used '''

    jsg.JSGLUE_JS_PATH = '/{0}/jsglue.js'.format(app_base)
    jsglue.init_app(app)
    flags.init_app(app)

    limiter.init_app(app)
    for handler in app.logger.handlers:
        limiter.logger.addHandler(handler)
    if app.config['RATELIMIT_ENABLED'] is False:
        limiter.enabled = False

    if app.config['USE_SENTRY']:
        sentry.init_app(app)

    # Initialize the Flask-Profiler ; see results at localhost:portnumber/app_base/flask-profiler
    if app.config['USE_PROFILER']:
        try:
            flask_profiler.init_app(app)
        except Exception as e:
            pass


def register_blueprints(app, url_prefix=None):
    ''' Register the Flask Blueprints used '''

    app.register_blueprint(index, url_prefix=url_prefix)
    app.register_blueprint(galaxy, url_prefix=url_prefix)
    app.register_blueprint(search, url_prefix=url_prefix)
    app.register_blueprint(plate, url_prefix=url_prefix)
    app.register_blueprint(images, url_prefix=url_prefix)
    app.register_blueprint(users, url_prefix=url_prefix)
    app.register_blueprint(jinjablue)
    app.register_blueprint(errors)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-07 13:54:18
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-27 11:19:28

from __future__ import print_function, division, absolute_import
from marvin.tests.api.conftest import ApiPage
import pytest


@pytest.mark.parametrize('page', [('api', 'CubeView:index')], ids=['cubes'], indirect=True)
class TestCubeView(object):

    def test_get_cube_success(self, page, params):
        page.load_page('get', page.url, params=params)
        data = 'this is a cube'
        page.assert_success(data)


@pytest.mark.parametrize('page', [('api', 'getCube')], ids=['getcubes'], indirect=True)
class TestGetCube(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_plateifu_success(self, galaxy, page, params, reqtype):
        data = {'plateifu': galaxy.plateifu, 'mangaid': galaxy.mangaid, 'ra': galaxy.ra, 'dec': galaxy.dec,
                'redshift': galaxy.redshift}
        page.load_page(reqtype, page.url.format(name=galaxy.plateifu), params=params)
        page.assert_success(data)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                       ('badname', 'name', 'String does not match expected pattern.'),
                                                       ('84', 'name', 'Shorter than minimum length 4.')],
                             ids=['norelease', 'badname', 'shortname'])
    def test_plateifu_failure(self, galaxy, page, reqtype, params, name, missing, errmsg):
        if name is None:
            page.route_no_valid_params(page.url.format(name=galaxy.plateifu), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(name=name), missing, reqtype=reqtype, params=params, errmsg=errmsg)


@pytest.mark.slow
@pytest.mark.parametrize('page', [('api', 'getCubeExtension')], ids=['getcubeext'], indirect=True)
class TestCubeExtension(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('cubeext', [('flux'), ('ivar'), ('mask')])
    def test_cubeext_success(self, galaxy, page, params, reqtype, cubeext):
        params.update({'name': galaxy.plateifu, 'cube_extension': cubeext})
        data = {'cube_extension': []}
        page.load_page(reqtype, page.url.format(**params), params=params)
        page.assert_success(data)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg, cubeext',
                             [(None, 'release', 'Missing data for required field.', 'flux'),
                              ('badname', 'name', 'String does not match expected pattern.', 'flux'),
                              ('84', 'name', 'Shorter than minimum length 4.', 'flux'),
                              ('8485-1901', 'cube_extension', 'Not a valid choice.', 'stuff')],
                             ids=['norelease', 'badname', 'shortname', 'badext'])
    def test_cubeext_failure(self, galaxy, page, reqtype, params, name, missing, errmsg, cubeext):
        params.update({'name': name, 'cube_extension': cubeext})
        if name is None:
            params.update({'name': galaxy.plateifu, 'cube_extension': cubeext})
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, params=params, errmsg=errmsg)




#!/usr/bin/env python
# encoding: utf-8
"""

test_modelGraph.py

Created by José Sánchez-Gallego on 24 Mar 2016.
Licensed under a 3-clause BSD license.

Revision history:
    24 Mar 2016 J. Sánchez-Gallego
      Initial version

"""

from __future__ import division
from __future__ import print_function

from brain.db.modelGraph import ModelGraph, nx

from sdss.internal.database.DatabaseConnection import DatabaseConnection

import unittest


# Assumes there is a local database named manga with mangadatadb and mangasampledb schemas.

DatabaseConnection('postgresql+psycopg2:///manga')

from sdss.internal.database.utah.mangadb import DataModelClasses, SampleModelClasses


class MarvinDB(object):
    datadb = DataModelClasses
    sampledb = SampleModelClasses

marvindb = MarvinDB()


class TestModelGraph(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Setups the test class."""

        cls.modelGraph = ModelGraph([marvindb.datadb, marvindb.sampledb])

    def testInitialitation(self):
        """Basic initialisation test."""

        # Initialises ModelGraph with datadb and sampledb
        modelGraphAll = ModelGraph([marvindb.datadb, marvindb.sampledb])

        testTables = ['mangasampledb.manga_target', 'mangasampledb.character',
                      'mangadatadb.cube', 'mangadatadb.fibers']
        for table in testTables:
            self.assertIn(table, modelGraphAll.nodes)

    def _testGetJoins(self, models, expected, nexus=None,
                      format_out='models'):
        """Convenience method to test getJoins()."""

        joins = self.modelGraph.getJoins(models, format_out=format_out,
                                         nexus=nexus)

        self.assertListEqual(joins, expected)

    def testGetJoins_oneModel(self):
        """Tests getJoins() with a single input model."""

        self._testGetJoins(marvindb.datadb.Cube, [marvindb.datadb.Cube])
        self._testGetJoins(marvindb.datadb.Cube, ['mangadatadb.cube'],
                           format_out='tables')
        self._testGetJoins('mangadatadb.cube', [marvindb.datadb.Cube])

    def testGetJoins_oneModel_nexus(self):
        """Tests getJoins() with a single input model and a nexus."""

        self._testGetJoins(marvindb.datadb.Cube, [], nexus=marvindb.datadb.Cube)
        self._testGetJoins(marvindb.datadb.IFUDesign, [marvindb.datadb.IFUDesign],
                           nexus=marvindb.datadb.Cube)
        self._testGetJoins(marvindb.datadb.Fibers, [marvindb.datadb.IFUDesign,
                                                    marvindb.datadb.Fibers],
                           nexus=marvindb.datadb.Cube)

    def testGetJoins_noPath(self):
        """Tests getJoins() when there is no path between two tables."""

        with self.assertRaises(nx.NetworkXNoPath):
            self._testGetJoins([marvindb.datadb.Cube, marvindb.datadb.MaskBit], [])

    def testGetJoin_fails(self):
        """Tests getJoins when input are wrong."""

        with self.assertRaises(ValueError):
            self._testGetJoins([1234], [])

        with self.assertRaises(ValueError):
            self._testGetJoins([], [])

        with self.assertRaises(AssertionError):
            self._testGetJoins([marvindb.datadb.Cube], [], format_out='bad_value')

        with self.assertRaises(AssertionError) as cm:
            self._testGetJoins([marvindb.datadb.Cube], [], nexus='bad_nexus')
            self.assertEqual(
                str(cm), 'nexus bad_nexus is not a node in the model graph')

    def testGetJoins_Models(self):
        """Tests getJoins() using a list of model classes."""

        models = [marvindb.datadb.Cube, marvindb.datadb.Fibers]
        expected = [marvindb.datadb.Cube, marvindb.datadb.IFUDesign, marvindb.datadb.Fibers]
        self._testGetJoins(models, expected)

        models = [marvindb.datadb.Cube, marvindb.datadb.PipelineCompletionStatus,
                  marvindb.datadb.FitsHeaderKeyword]
        expected = [marvindb.datadb.Cube, marvindb.datadb.PipelineInfo,
                    marvindb.datadb.PipelineCompletionStatus,
                    marvindb.datadb.FitsHeaderValue, marvindb.datadb.FitsHeaderKeyword]
        self._testGetJoins(models, expected)

        models = [marvindb.datadb.Fibers, marvindb.sampledb.MangaTarget]
        expected = ['mangadatadb.fibers', 'mangadatadb.ifudesign',
                    'mangadatadb.cube', 'mangasampledb.manga_target']
        self._testGetJoins(models, expected, format_out='tables')

        models = [marvindb.datadb.Cube, marvindb.datadb.Fibers, marvindb.datadb.Cube]
        expected = [u'mangadatadb.cube', u'mangadatadb.ifudesign',
                    u'mangadatadb.fibers']
        self._testGetJoins(models, expected, format_out='tables')

    def testGetJoins_Models_nexus(self):
        """Tests getJoins() using a list of model classes and a nexus."""

        nexus = marvindb.datadb.Cube
        models = [marvindb.datadb.Cube, marvindb.datadb.Fibers, marvindb.datadb.FitsHeaderKeyword]
        expected = [marvindb.datadb.IFUDesign, marvindb.datadb.Fibers,
                    marvindb.datadb.FitsHeaderValue, marvindb.datadb.FitsHeaderKeyword]
        expected_tables = ['mangadatadb.ifudesign', 'mangadatadb.fibers',
                           'mangadatadb.fits_header_value',
                           'mangadatadb.fits_header_keyword']

        self._testGetJoins(models, expected, nexus=nexus)
        self._testGetJoins(models, expected_tables, format_out='tables',
                           nexus=nexus)


if __name__ == '__main__':
    # set to 1 for the usual '...F..' style output, or 2 for more verbose output.
    verbosity = 2
    unittest.main(verbosity=verbosity)



# License information goes here
# -*- coding: utf-8 -*-
"""
===========
sdss_access
===========

The function sdss4tools.install.version() should be used to set the ``__version__``
package variable.  In order for this to work properly, the svn property
svn:keywords must be set to HeadURL on this file.

.. SDSS-IV: http://trac.sdss.org
.. Python:  http://python.org
"""
#
from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.
#from sdss4tools.install import version
#
# Set version string.
#
#__version__ = version('$HeadURL$')
#
# Clean up namespace
#
#del version
__version__ = "trunk"

from .path import Path as SDSSPath, AccessError
from .sync import HttpAccess, RsyncAccess



#!/usr/bin/env python2
# encoding: utf-8
#
# cube.py


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import warnings

from astropy.io import fits
from astropy.wcs import WCS
import numpy as np

import marvin
import marvin.core.exceptions
import marvin.tools.spaxel
import marvin.tools.maps
import marvin.utils.general.general
from marvin.utils.general import get_nsa_data

from marvin.core.core import MarvinToolsClass
from marvin.core.exceptions import MarvinError, MarvinUserWarning

try:
    import photutils.aperture_funcs
except ImportError:
    photutils = False


class Cube(MarvinToolsClass):
    """A class to interface with MaNGA DRP data cubes.

    This class represents a fully reduced DRP data cube, initialised either
    from a file, a database, or remotely via the Marvin API.

    Parameters:
        data (``HDUList``, SQLAlchemy object, or None):
            An astropy ``HDUList`` or a SQLAlchemy object of a cube, to
            be used for initialisation. If ``None``, the normal mode will
            be used (see :ref:`mode-decision-tree`).
        filename (str):
            The path of the file containing the data cube to load.
        mangaid (str):
            The mangaid of the data cube to load.
        plateifu (str):
            The plate-ifu of the data cube to load (either ``mangaid`` or
            ``plateifu`` can be used, but not both).
        nsa_source ({'auto', 'drpall', 'nsa'}):
            Defines how the NSA data for this object should loaded when
            ``Cube.nsa`` is first called. If ``drpall``, the drpall file will
            be used (note that this will only contain a subset of all the NSA
            information); if ``nsa``, the full set of data from the DB will be
            retrieved. If the drpall file or a database are not available, a
            remote API call will be attempted. If ``nsa_source='auto'``, the
            source will depend on how the ``Cube`` object has been
            instantiated. If the cube has ``Cube.data_origin='file'``,
            the drpall file will be used (as it is more likely that the user
            has that file in their system). Otherwise, ``nsa_source='nsa'``
            will be assumed. This behaviour can be modified during runtime by
            modifying the ``Cube.nsa_mode`` with one of the valid values.
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See :ref:`mode-decision-tree`.
        release (str):
            The MPL/DR version of the data to use.

    Return:
        cube:
            An object representing the data cube.

    """

    def __init__(self, *args, **kwargs):

        valid_kwargs = [
            'data', 'filename', 'mangaid', 'plateifu', 'mode', 'release',
            'drpall', 'nsa_source']

        assert len(args) == 0, 'Cube does not accept arguments, only keywords.'
        for kw in kwargs:
            assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

        self.shape = None
        self.wcs = None
        self.wavelength = None
        self._drpall_data = None

        self._flux = None
        self._ivar = None
        self._mask = None

        super(Cube, self).__init__(*args, **kwargs)

        if self.data_origin == 'file':
            self._load_cube_from_file(data=self.data)
        elif self.data_origin == 'db':
            self._load_cube_from_db(data=self.data)
        elif self.data_origin == 'api':
            self._load_cube_from_api()

        self._init_attributes()

        # Checks that the drpver set in MarvinToolsClass matches the header
        header_drpver = self.header['VERSDRP3'].strip()
        header_drpver = 'v1_5_1' if header_drpver == 'v1_5_0' else header_drpver
        assert header_drpver == self._drpver, ('mismatch between cube._drpver={0} '
                                               'and header drpver={1}'.format(self._drpver,
                                                                              header_drpver))

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(Cube, self)._getFullPath('mangacube', ifu=ifu,
                                              drpver=self._drpver, plate=plate)

    def download(self):
        """Downloads the cube using sdss_access - Rsync,"""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(Cube, self).download('mangacube', ifu=ifu,
                                          drpver=self._drpver, plate=plate)

    def __repr__(self):
        """Representation for Cube."""

        return ('<Marvin Cube (plateifu={0}, mode={1}, data_origin={2})>'
                .format(repr(self.plateifu), repr(self.mode),
                        repr(self.data_origin)))

    def __getitem__(self, xy):
        """Returns the spaxel for ``(x, y)``"""

        return self.getSpaxel(x=xy[1], y=xy[0], xyorig='lower')

    def _init_attributes(self):
        """Initialises several attributes."""

        self.ra = float(self.header['OBJRA'])
        self.dec = float(self.header['OBJDEC'])

        self.plate = int(self.header['PLATEID'])
        self.ifu = int(self.header['IFUDSGN'])
        self.mangaid = self.header['MANGAID']

        self._isbright = 'APOGEE' in self.header['SRVYMODE']

        self.dir3d = 'mastar' if self._isbright else 'stack'

    def _load_cube_from_file(self, data=None):
        """Initialises a cube from a file."""
        if data is not None:
            assert isinstance(data, fits.HDUList), 'data is not an HDUList object'
        else:
            try:
                self.data = fits.open(self.filename)
            except IOError as err:
                raise IOError('filename {0} cannot be found: {1}'.format(self.filename, err))

        self.header = self.data[1].header
        self.shape = self.data['FLUX'].data.shape[1:]
        self.wcs = WCS(self.header)
        self.wavelength = self.data['WAVE'].data
        self.plateifu = self.header['PLATEIFU']

        # Checks and populates the release.
        file_drpver = self.header['VERSDRP3']
        file_drpver = 'v1_5_1' if file_drpver == 'v1_5_0' else file_drpver

        file_ver = marvin.config.lookUpRelease(file_drpver)
        assert file_ver is not None, 'cannot find file version.'

        if file_ver != self._release:
            warnings.warn('mismatch between file release={0} and object release={1}. '
                          'Setting object release to {0}'.format(file_ver, self._release),
                          MarvinUserWarning)
            self._release = file_ver

            # Reload NSA data from file version of drpall file
            self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)
            self._drpall = marvin.config._getDrpAllPath(file_drpver)
            self.mangaid = self.header['MANGAID']

            nsa_source = 'drpall' if self.nsa_source == 'auto' else self.nsa_source

            self._nsa = None
            self._nsa = get_nsa_data(self.mangaid, mode='auto', source=nsa_source,
                                     drpver=self._drpver, drpall=self._drpall)

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

    def _load_cube_from_db(self, data=None):
        """Initialises a cube from the DB."""
        mdb = marvin.marvindb
        plate, ifu = self.plateifu.split('-')

        if not mdb.isdbconnected:
            raise MarvinError('No db connected')
        else:
            import sqlalchemy
            datadb = mdb.datadb

            if self.data:
                assert isinstance(data, datadb.Cube), 'data is not an instance of mangadb.Cube.'
                self.data = data
            else:
                try:
                    self.data = mdb.session.query(datadb.Cube).join(
                        datadb.PipelineInfo,
                        datadb.PipelineVersion,
                        datadb.IFUDesign).filter(
                            mdb.datadb.PipelineVersion.version == self._drpver,
                            datadb.Cube.plate == int(plate), datadb.IFUDesign.name == ifu).one()
                except sqlalchemy.orm.exc.MultipleResultsFound as ee:
                    raise MarvinError('Could not retrieve cube for plate-ifu {0}: '
                                      'Multiple Results Found: {1}'.format(self.plateifu, ee))
                except sqlalchemy.orm.exc.NoResultFound as ee:
                    raise MarvinError('Could not retrieve cube for plate-ifu {0}: '
                                      'No Results Found: {1}'.format(self.plateifu, ee))
                except Exception as ee:
                    raise MarvinError('Could not retrieve cube for plate-ifu {0}: '
                                      'Unknown exception: {1}'.format(self.plateifu, ee))

            self.header = self.data.header
            self.wcs = WCS(self.data.wcs.makeHeader())
            self.data = self.data
            self.shape = self.data.shape.shape
            self.wavelength = self.data.wavelength.wavelength

    def _load_cube_from_api(self):
        """Calls the API and retrieves the necessary information to instantiate the cube."""
        url = marvin.config.urlmap['api']['getCube']['url']

        try:
            response = self._toolInteraction(url.format(name=self.plateifu))
        except Exception as ee:
            raise MarvinError('found a problem when checking if remote cube '
                              'exists: {0}'.format(str(ee)))

        data = response.getData()

        self.header = fits.Header.fromstring(data['header'])
        self.shape = data['shape']
        self.wavelength = np.array(data['wavelength'])
        self.wcs = WCS(fits.Header.fromstring(data['wcs_header']))

        if self.plateifu != data['plateifu']:
            raise MarvinError('remote cube has a different plateifu!')

        return

    def _getExtensionData(self, extName):
        """Returns the data from an extension."""

        if self.data_origin == 'file':
            return self.data[extName.upper()].data

        elif self.data_origin == 'db':
            return self.data.get3DCube(extName.lower())

        elif self.data_origin == 'api':

            url = marvin.config.urlmap['api']['getCubeExtension']['url']

            try:
                response = self._toolInteraction(url.format(name=self.plateifu,
                                                            cube_extension=extName.lower()))
            except Exception as ee:
                raise MarvinError('found a problem when checking if remote cube '
                                  'exists: {0}'.format(str(ee)))

            data = response.getData()

            return np.array(data['cube_extension'])

    @property
    def flux(self):
        """Gets the ``FLUX`` data extension."""
        if self._flux is None:
            self._flux = self._getExtensionData('FLUX')
        return self._flux

    @property
    def ivar(self):
        """Gets the ``IVAR`` data extension."""
        if self._ivar is None:
            self._ivar = self._getExtensionData('IVAR')
        return self._ivar

    @property
    def mask(self):
        """Gets the ``MASK`` data extension."""
        if self._mask is None:
            self._mask = self._getExtensionData('MASK')
        return self._mask

    @property
    def qualitybit(self):
        """The Cube DRP3QUAL bits."""

        # Python 2-3 compatibility
        try:
            bit = long(self.header['DRP3QUAL'])
        except NameError:
            bit = int(self.header['DRP3QUAL'])

        labels = None

        # get labels
        if self.data_origin == 'db':
            labels = self.data.getQualFlags()
        elif self.data_origin == 'file':
            pass
        elif self.data_origin == 'api':
            pass

        return 'DRP3QUAL', bit, labels

    @property
    def targetbit(self):
        """The Cube MNGTRG bits."""

        try:
            names = ['MNGTARG1', 'MNGTARG2', 'MNGTARG3']
            targs = [int(self.header[names[0]]), int(self.header[names[1]]),
                     int(self.header[names[2]])]
        except KeyError:
            names = ['MNGTRG1', 'MNGTRG2', 'MNGTRG3']
            targs = [int(self.header[names[0]]), int(self.header[names[1]]),
                     int(self.header[names[2]])]

        ind = np.nonzero(targs)[0]

        finaltargs = {}
        finaltargs['names'] = [names[i] for i in ind]
        finaltargs['bits'] = [targs[i] for i in ind]

        # get labels
        if self.data_origin == 'db':
            finaltargs['labels'] = [self.data.getTargFlags(type=i + 1) for i in ind]
        elif self.data_origin == 'file':
            pass
        elif self.data_origin == 'api':
            pass

        return finaltargs

    def getSpaxel(self, x=None, y=None, ra=None, dec=None,
                  properties=True, modelcube=False, **kwargs):
        """Returns the |spaxel| matching certain coordinates.

        The coordinates of the spaxel to return can be input as ``x, y`` pixels
        relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
        coordinates.

        Parameters:
            x,y (int or array):
                The spaxel coordinates relative to ``xyorig``. If ``x`` is an
                array of coordinates, the size of ``x`` must much that of
                ``y``.
            ra,dec (float or array):
                The coordinates of the spaxel to return. The closest spaxel to
                those coordinates will be returned. If ``ra`` is an array of
                coordinates, the size of ``ra`` must much that of ``dec``.
            xyorig ({'center', 'lower'}):
                The reference point from which ``x`` and ``y`` are measured.
                Valid values are ``'center'``, for the centre of the
                spatial dimensions of the cube, or ``'lower'`` for the
                lower-left corner. This keyword is ignored if ``ra`` and
                ``dec`` are defined. ``xyorig`` defaults to
                ``marvin.config.xyorig.``
            properties (bool):
                If ``True``, the spaxel will be initiated with the DAP
                properties from the default Maps matching this cube.
            modelcube (:class:`~marvin.tools.modelcube.ModelCube` or None or bool):
                A :class:`~marvin.tools.modelcube.ModelCube` object
                representing the DAP modelcube entity. If None, the |spaxel|
                will be returned without model information. Default is False.

        Returns:
            spaxels (list):
                The |spaxel| objects for this cube corresponding to the input
                coordinates. The length of the list is equal to the number
                of input coordinates.

        .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

        """

        # TODO: do we want to use x/y, ra/dec, or a single coords parameter (as
        # an array of coordinates) and a mode keyword.

        kwargs['cube'] = self
        kwargs['maps'] = properties
        kwargs['modelcube'] = modelcube

        return marvin.utils.general.general.getSpaxel(x=x, y=y, ra=ra, dec=dec, **kwargs)

    def getMaps(self, **kwargs):
        """Retrieves the DAP :class:`~marvin.tools.maps.Maps` for this cube.

        If called without additional ``kwargs``, :func:`getMaps` will initilise
        the :class:`~marvin.tools.maps.Maps` using the ``plateifu`` of this
        :class:`~marvin.tools.cube.Cube`. Otherwise, the ``kwargs`` will be
        passed when initialising the :class:`~marvin.tools.maps.Maps`.

        """

        if len(kwargs.keys()) == 0 or 'filename' not in kwargs:
            kwargs.update({'plateifu': self.plateifu, 'release': self._release})

        maps = marvin.tools.maps.Maps(**kwargs)
        maps._cube = self
        return maps

    def getAperture(self, coords, radius, mode='pix', weight=True,
                    return_type='mask'):
        """Returns the spaxel in a circular or elliptical aperture.

        Returns either a mask of the same shape as the cube with the spaxels
        within an aperture, or the integrated spaxel from combining the spectra
        for those spaxels.

        The centre of the aperture is defined by ``coords``, which must be a
        tuple of ``(x,y)`` (if ``mode='pix'``) or ``(ra,dec)`` coordinates
        (if ``mode='sky'``). ``radius`` defines the radius of the circular
        aperture, or the parameters of the aperture ellipse.

        If ``weight=True``, the returned mask indicated the fraction of the
        spaxel encompassed by the aperture, ranging from 0 for spaxels not
        included to 1 for pixels totally included in the aperture. This
        weighting is used to return the integrated spaxel.

        Parameters:
            coords (tuple):
                Either the ``(x,y)`` or ``(ra,dec)`` coordinates of the centre
                of the aperture.
            radius (float or tuple):
                If a float, the radius of the circular aperture. If
                ``mode='pix'`` it must be the radius in pixels; if
                ``mode='sky'``, ``radius`` is in arcsec. To define an
                elliptical aperture, ``radius`` must be a 3-element tuple with
                the first two elements defining the major and minor semi-axis
                of the ellipse, and the third one the position angle in degrees
                from North to East.
            mode ({'pix', 'sky'}):
                Defines whether the values in ``coords`` and ``radius`` refer
                to pixels in the cube or angles on the sky.
            weight (bool):
                If ``True``, the returned mask or integrated spaxel will be
                weighted by the fractional pixels in the aperture.
            return_type ({'mask', 'mean', 'median', 'sum', 'spaxels'}):
                The type of data to be returned.

        Returns:
            result:
                If ``return_type='mask'``, this methods returns a 2D mask with
                the shape of the cube indicating the spaxels included in the
                aperture and, if appliable, their fractional contribution to
                the aperture. If ``spaxels``, both the mask (flattened to a
                1D array) and the :class:`~marvin.tools.spaxel.Spaxel`
                included in the aperture are returned. ``mean``, ``median``,
                or ``sum`` will allow arithmetic operations with the spaxels
                in the aperture in the future.

        Example:
            To get the mask for a circular aperture centred in spaxel (5, 7)
            and with radius 5 spaxels

                >>> mask = cube.getAperture((5, 7), 5)
                >>> mask.shape
                (34, 34)

            If you want to get the spaxels associated with that mask

                >>> mask, spaxels = cube.getAperture((5, 7), 5, return_type='spaxels')
                >>> len(spaxels)
                15
        """

        assert return_type in ['mask', 'mean', 'median', 'sum', 'spaxels']

        if return_type not in ['mask', 'spaxels']:
            raise marvin.core.exceptions.MarvinNotImplemented(
                'return_type={0} is not yet implemented'.format(return_type))

        if not photutils:
            raise MarvinError('getAperture currently requires photutils.')

        if mode != 'pix':
            raise marvin.core.exceptions.MarvinNotImplemented(
                'mode={0} is not yet implemented'.format(mode))

        if not np.isscalar(radius):
            raise marvin.core.exceptions.MarvinNotImplemented(
                'elliptical apertures are not yet implemented')

        data_mask = np.zeros(self.shape)

        if weight:
            phot_mode = ''
        else:
            phot_mode = 'center'

        coords = np.atleast_2d(coords) + 0.5

        mask = photutils.aperture_funcs.get_circular_fractions(
            data_mask, coords, radius, phot_mode, 0)

        if return_type == 'mask':
            return mask

        if return_type == 'spaxels':
            mask_idx = np.where(mask)
            spaxels = self.getSpaxel(x=mask_idx[0], y=mask_idx[1],
                                     xyorig='lower')

            fractions = mask[mask_idx]

            return (fractions, spaxels)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-10-13 04:13:10
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2016-11-29 17:13:34

from __future__ import print_function, division, absolute_import




#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-26 10:42:12
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-26 10:42:12 by Brian Cherinka
    Last Modified On: 2016-04-26 10:42:12 by Brian

'''
from __future__ import print_function, division
from flask import Blueprint, render_template, session as current_session, request, jsonify, url_for
from flask_classy import route
from brain.api.base import processRequest
from marvin.core.exceptions import MarvinError
from marvin.tools.query import doQuery, Query
from marvin.tools.query.forms import MarvinForm
from marvin.web.controllers import BaseWebView
from marvin.api.base import arg_validate as av
from marvin.tools.query.query_utils import query_params, bestparams
from wtforms import validators, ValidationError
from marvin.utils.general import getImagesByList
from marvin.web.web_utils import buildImageDict
from marvin.web.extensions import limiter
import random

search = Blueprint("search_page", __name__)


def getRandomQuery():
    ''' Return a random query from this list '''
    samples = ['nsa.z < 0.02', 'cube.plate < 8000', 'haflux > 25',
               'nsa.sersic_logmass > 9.5 and nsa.sersic_logmass < 11', 'emline_ew_ha_6564 > 3']
    q = random.choice(samples)
    return q


def all_in(fullist):
    ''' search form query parameter form validation '''
    def _all_in(form, field):
        myparams = [f for f in field.data if f]
        outsiders = list(set(myparams) - set(fullist))
        message = '{0} must be a given query parameter'.format(', '.join(outsiders))
        if outsiders:
            raise ValidationError(message)
    return _all_in


class Search(BaseWebView):
    route_base = '/search/'

    def __init__(self):
        ''' Initialize the route '''
        super(Search, self).__init__('marvin-search')
        self.search = self.base.copy()
        self.search['filter'] = None
        self.search['results'] = None
        self.search['errmsg'] = None
        self.search['returnparams'] = None
        self.mf = MarvinForm()

    def before_request(self, *args, **kwargs):
        ''' Do these things before a request to any route '''
        super(Search, self).before_request(*args, **kwargs)
        self.reset_dict(self.search)

    @route('/', methods=['GET', 'POST'])
    @limiter.limit("60/minute")
    def index(self):

        # Attempt to retrieve search parameters
        form = processRequest(request=request)
        self.search['formparams'] = form

        # set the search form and form validation
        searchform = self.mf.SearchForm(form)
        searchform.returnparams.choices = [(k.lower(), k) for k in query_params.list_params()]

        # Add the forms and parameters
        self.search['paramdata'] = query_params
        self.search['guideparams'] = [{'id': p.full, 'optgroup': group.name, 'type': 'double' if p.dtype == 'float' else p.dtype, 'validation': {'step': 'any'}} for group in query_params for p in group]
        self.search['searchform'] = searchform
        self.search['placeholder'] = getRandomQuery()

        # If form parameters then try to do a search
        if form:
            self.search.update({'results': None, 'errmsg': None})

            args = av.manual_parse(self, request, use_params='search')
            # get form parameters
            searchvalue = form['searchbox']  # search filter input
            returnparams = form.getlist('returnparams', type=str)  # dropdown select
            self.search.update({'returnparams': returnparams})
            current_session.update({'searchvalue': searchvalue, 'returnparams': returnparams})

            # if main form passes validation then do search
            if searchform.validate():
                # try the query
                try:
                    q, res = doQuery(searchfilter=searchvalue, release=self._release, returnparams=returnparams)
                except MarvinError as e:
                    self.search['errmsg'] = 'Could not perform query: {0}'.format(e)
                else:
                    self.search['filter'] = q.strfilter
                    self.search['count'] = res.totalcount
                    self.search['runtime'] = res.query_runtime.total_seconds()
                    if res.count > 0:
                        cols = res.mapColumnsToParams()
                        rows = res.getDictOf(format_type='listdict')
                        output = {'total': res.totalcount, 'rows': rows, 'columns': cols, 'limit': None, 'offset': None}
                    else:
                        output = None
                    self.search['results'] = output
                    self.search['reslen'] = len(res.results)
                    if returnparams:
                        returnparams = [str(r) for r in returnparams]
                    rpstr = 'returnparams={0} <br>'.format(returnparams) if returnparams else ''
                    qstr = ', returnparams=returnparams' if returnparams else ''
                    self.search['querystring'] = ("<html><samp>from marvin.tools.query import Query<br>\
                        filter='{0}'<br> {1}\
                        q = Query(searchfilter=filter{2})<br>\
                        r = q.run()<br></samp></html>".format(searchvalue, rpstr, qstr))

        return render_template('search.html', **self.search)

    @route('/dopost/', methods=['POST'], endpoint='dopost')
    def post(self):
        return 'this is a post'

    @route('/getparams/<paramdisplay>/', methods=['GET', 'POST'], endpoint='getparams')
    def getparams(self, paramdisplay):
        ''' Retrieves the list of query parameters for Bloodhound Typeahead

        '''

        # set the paramdisplay if it is not
        if not paramdisplay:
            paramdisplay = 'best'

        # run query and retrieve parameters
        q = Query(release=self._release)
        if paramdisplay == 'all':
            params = q.get_available_params('all')
        elif paramdisplay == 'best':
            params = bestparams
        output = jsonify(params)
        return output

    @route('/webtable/', methods=['GET', 'POST'], endpoint='webtable')
    def webtable(self):
        ''' Do a query for Bootstrap Table interaction in Marvin web '''

        form = processRequest(request=request)
        args = av.manual_parse(self, request, use_params='query')

        #{'sort': u'cube.mangaid', 'task': None, 'end': None, 'searchfilter': None,
        #'paramdisplay': None, 'start': None, 'rettype': None, 'limit': 10, 'offset': 30,
        #'release': u'MPL-4', 'params': None, 'order': u'asc'}

        # remove args
        __tmp__ = args.pop('release', None)
        __tmp__ = args.pop('searchfilter', None)
        limit = args.get('limit')
        offset = args.get('offset')
        if 'sort' in args:
            current_session['query_sort'] = args['sort']

        # set parameters
        searchvalue = current_session.get('searchvalue', None)
        returnparams = current_session.get('returnparams', None)
        # limit = form.get('limit', 10, type=int)
        # offset = form.get('offset', None, type=int)
        # order = form.get('order', None, type=str)
        # sort = form.get('sort', None, type=str)
        # search = form.get('search', None, type=str)

        # exit if no searchvalue is found
        if not searchvalue:
            output = jsonify({'webtable_error': 'No searchvalue found', 'status': -1})
            return output

        # do query
        q, res = doQuery(searchfilter=searchvalue, release=self._release, returnparams=returnparams, **args)
        # q, res = doQuery(searchfilter=searchvalue, release=self._release,
        #                  limit=limit, order=order, sort=sort, returnparams=returnparams)
        # get subset on a given page
        __results__ = res.getSubset(offset, limit=limit)
        # get keys
        cols = res.mapColumnsToParams()
        # create output
        rows = res.getDictOf(format_type='listdict')
        output = {'total': res.totalcount, 'rows': rows, 'columns': cols, 'limit': limit, 'offset': offset}
        output = jsonify(output)
        return output

    @route('/postage/', methods=['GET', 'POST'], defaults={'page': 1}, endpoint='postage')
    @route('/postage/<page>/', methods=['GET', 'POST'], endpoint='postage')
    def postagestamp(self, page):
        ''' Get the postage stamps for a set of query results in the web '''

        postage = {}
        postage['error'] = None
        postage['images'] = None

        pagesize = 16  # number of rows (images) in a page
        pagenum = int(page)  # current page number
        searchvalue = current_session.get('searchvalue', None)
        if not searchvalue:
            postage['error'] = 'No query found! Cannot generate images without a query.  Go to the Query Page!'
            return render_template('postage.html', **postage)

        sort = current_session.get('query_sort', 'cube.mangaid')
        offset = (pagesize * pagenum) - pagesize
        q, res = doQuery(searchfilter=searchvalue, release=self._release, sort=sort, limit=10000)
        plateifus = res.getListOf('plateifu')
        # if a dap query, grab the unique galaxies
        if q._isdapquery:
            plateifus = list(set(plateifus))

        # only grab subset if more than 16 galaxies
        if len(plateifus) > pagesize:
            # __results__ = res.getSubset(offset, limit=pagesize)
            # plateifus = res.getListOf('plateifu')
            plateifus = plateifus[offset:offset + pagesize]

        # get images
        imfiles = None
        try:
            imfiles = getImagesByList(plateifus, as_url=True, mode='local', release=self._release)
        except MarvinError as e:
            postage['error'] = 'Error: could not get images: {0}'.format(e)
        else:
            images = buildImageDict(imfiles)

        # if image grab failed, make placeholders
        if not imfiles:
            images = buildImageDict(imfiles, test=True, num=pagesize)

        # Compute page stats
        totalpages = int(res.totalcount // pagesize) + int(res.totalcount % pagesize != 0)
        page = {'size': pagesize, 'active': int(page), 'total': totalpages, 'count': res.totalcount}

        postage['page'] = page
        postage['images'] = images
        return render_template('postage.html', **postage)


Search.register(search)




# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

"""
SQLAlchemy-boolean-search
=========================
SQLAlchemy-boolean-search translates a boolean search expression such as::

    field1=*something* and not (field2==1 or field3<=10.0)

into its corresponding SQLAlchemy query filter.

Install
-------

    pip install sqlalchemy-boolean-search

Usage example
-------------

    from sqlalchemy_boolean_search import parse_boolean_search

    # DataModel defined elsewhere (with field1, field2 and field3)
    from app.models import DataModel

    # Parse boolean search into a parsed expression
    boolean_search = 'field1=*something* and not (field2==1 or field3<=10.0)'
    parsed_expression = parse_boolean_search(boolean_search)

    # Retrieve records using a filter generated by the parsed expression
    records = DataModel.query.filter(parsed_expression.filter(DataModel))

Documentation
-------------
http://sqlalchemy-boolean-search.readthedocs.org/

Authors
-------
* Ling Thio - ling.thio [at] gmail.com

Revision History
--------
2016-03-5: Modified to allow for a list of ModelClasses as input - Brian Cherinka
2016-03-11: Modified to output a dictionary of parameters: values - B. Cherinka
2016-03-16: Changed sqlalchemy values in conditions to bindparam for post-replacement - B. Cherinka
2016-03-24: Allowed for = to mean equality for non string fields and LIKE for strings - B. Cherinka
          : Changed the dot relationship in get_field to filter on the relationship_name first - B. Cherinka
2016-05-11: Added support for PostgreSQL array filters, using value op ANY(array) - B. Cherinka
2016-09-12: Modified to allow for function names (with nested expression) in the expression - B. Cherinka
2016-09-12: Modified to separate out function conditions from regular conditions
2016-09-21: Modified field checks to allow for hybrid properties to pass through - B. Cherinka
2016-09-24: Added in Decimal field as an fieldtype option - B. Cherinka
2016-09-29: Modified the bindparam name to allow for value ranges of a single name - B. Cherinka
"""

from __future__ import print_function
import pyparsing as pp
import inspect
import decimal
from pyparsing import ParseException  # explicit export
from sqlalchemy import func, bindparam
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import or_, and_, not_, sqltypes, between
from operator import le, ge, gt, lt, eq, ne

opdict = {'<=': le, '>=': ge, '>': gt, '<': lt, '!=': ne, '==': eq, '=': eq}


# Define a custom exception class
class BooleanSearchException(Exception):
    pass


# ***** Utility functions *****
def get_field(DataModelClass, field_name, base_name=None):
    """ Returns a SQLAlchemy Field from a field name such as 'name' or 'parent.name'.
        Returns None if no field exists by that field name.
    """
    # Handle hierarchical field names such as 'parent.name'
    if base_name:
        if base_name in DataModelClass.__tablename__:
            return getattr(DataModelClass, field_name, None)
        else:
            return None

    # Handle flat field names such as 'name'
    return getattr(DataModelClass, field_name, None)


# ***** Define the expression element classes *****

class FxnCondition(object):
    ''' Represents a fxn-operand-value condition where
    the fxn is a function containing a new condition
    '''
    def __init__(self, data):
        self.fxn = data[0][0]
        self.fxnname = self.fxn[0]
        self.fxncond = self.fxn[1]
        self.op = data[0][1]
        self.value = data[0][2]

    def filter(self, DataModelClass):
        return None

    def __repr__(self):
        return '{0}({1})'.format(self.fxnname, self.fxncond) + self.op + self.value


class Condition(object):
    """ Represents a 'name operand value' condition,
        where operand can be one of: '<', '<=', '=', '==', '!=', '>=', '>'.
    """
    def __init__(self, data):
        self.fullname = data[0][0]
        if '.' in self.fullname:
            self.basename, self.name = self.fullname.split('.', 1)
        else:
            self.basename = None
            self.name = self.fullname
        self.op = data[0][1]
        self.value = data[0][2]
        if self.op == 'between':
            self.value2 = data[0][4]
        uniqueparams.append(self.fullname)
        if self.fullname not in params:
            params.update({self.fullname: self.value})
            self.bindname = self.fullname
        else:
            count = list(params.keys()).count(self.fullname)
            self.bindname = '{0}_{1}'.format(self.fullname, count)
            params.update({self.fullname: self.value})

    def filter(self, DataModelClass):
        ''' Return the condition as an SQLalchemy query condition '''

        condition = None
        if inspect.ismodule(DataModelClass):
            # one module
            models = [i[1] for i in inspect.getmembers(DataModelClass, inspect.isclass) if hasattr(i[1], '__tablename__')]
        else:
            # list of Model Classes
            if isinstance(DataModelClass, list):
                models = DataModelClass
            else:
                models = None

        if models:
            # Input is a a list of DataModelClasses
            field = None
            index = None
            for i, model in enumerate(models):

                field = get_field(model, self.name, base_name=self.basename)
                try:
                    ptype = field.type
                    ilike = field.ilike
                except AttributeError as e:
                    ptype = None
                    ilike = None

                if not isinstance(field, type(None)) and ptype and ilike:
                    index = i
                    break

            if isinstance(field, type(None)):
                raise BooleanSearchException(
                    "Table '%(table_name)s' does not have a field named '%(field_name)s'."
                    % dict(table_name=model.__tablename__, field_name=self.name))

            condition = self.filter_one(models[index], field=field, condition=condition)

        else:
            # Input is only one DataModelClass
            field = get_field(DataModelClass, self.name)
            if field:
                condition = self.filter_one(DataModelClass, field=field, condition=condition)
            else:
                raise BooleanSearchException(
                    "Table '%(table_name)s' does not have a field named '%(field_name)s'."
                    % dict(table_name=DataModelClass.__tablename__, field_name=self.name))

        return condition

    def format_value(self, value, fieldtype, field):
        ''' Formats the value based on the fieldtype '''

        if fieldtype == float or fieldtype == decimal.Decimal:
            try:
                outvalue = float(value)
                lower_field = field
            except:
                raise BooleanSearchException(
                    "Field {0} expects a float value. Received value {1} instead.".format(self.name, value))
        elif fieldtype == int:
            try:
                outvalue = int(value)
                lower_field = field
            except:
                raise BooleanSearchException(
                    "Field {0} expects an integer value. Received value {1} instead.".format(self.name, value))
        else:
            lower_field = func.lower(field)
            outvalue = value

        return outvalue, lower_field

    def bindAndLowerValue(self, field):
        '''Bind and lower the value based on field type '''

        lower_value_2 = None
        # get python field type
        ftypes = [float, int, decimal.Decimal]
        fieldtype = field.type.python_type

        value, lower_field = self.format_value(self.value, fieldtype, field)
        if hasattr(self, 'value2'):
            value2, lower_field = self.format_value(self.value2, fieldtype, field)

        # Bind the parameter value to the parameter name
        boundvalue = bindparam(self.bindname, value)
        lower_value = func.lower(boundvalue) if fieldtype not in ftypes else boundvalue
        if hasattr(self, 'value2'):
            self.bindname = '{0}_{1}'.format(self.fullname, 2)
            boundvalue2 = bindparam(self.bindname, value2)
            lower_value_2 = func.lower(boundvalue2) if fieldtype not in ftypes else boundvalue2

        return lower_field, lower_value, lower_value_2

    def filter_one(self, DataModelClass, field=None, condition=None):
        """ Return the condition as a SQLAlchemy query condition
        """
        if not isinstance(field, type(None)):
            # Prepare field and value
            lower_field, lower_value, lower_value_2 = self.bindAndLowerValue(field)

            # Handle Arrays
            if isinstance(field.type, postgresql.ARRAY):
                condition = field.any(self.value, operator=opdict[self.op])
            else:
                # Do Normal Scalar Stuff

                # Return SQLAlchemy condition based on operator value
                # self.name is parameter name, lower_field is Table.parameterName
                if self.op == '==':
                    condition = lower_field.__eq__(lower_value)
                elif self.op == '<':
                    condition = lower_field.__lt__(lower_value)
                elif self.op == '<=':
                    condition = lower_field.__le__(lower_value)
                elif self.op == '>':
                    condition = lower_field.__gt__(lower_value)
                elif self.op == '>=':
                    condition = lower_field.__ge__(lower_value)
                elif self.op == '!=':
                    condition = lower_field.__ne__(lower_value)
                elif self.op == '=':
                    if isinstance(field.type, sqltypes.TEXT) or \
                       isinstance(field.type, sqltypes.VARCHAR) or \
                       isinstance(field.type, sqltypes.String):
                        # this operator maps to LIKE
                        # x=5 -> x LIKE '%5%' (x contains 5)
                        # x=5* -> x LIKE '5%' (x starts with 5)
                        # x=*5 -> x LIKE '%5' (x ends with 5)
                        field = getattr(DataModelClass, self.name)
                        value = self.value
                        if value.find('*') >= 0:
                            value = value.replace('*', '%')
                            condition = field.ilike(bindparam(self.bindname, value))
                        else:
                            condition = field.ilike('%' + bindparam(self.bindname, value) + '%')
                    else:
                        # if not a text column, then use "=" as a straight equals
                        condition = lower_field.__eq__(lower_value)
                elif self.op == 'between':
                    condition = between(lower_field, lower_value, lower_value_2)

        return condition

    def __repr__(self):
        more = 'and' + self.value2 if hasattr(self, 'value2') else ''
        return self.fullname + self.op + self.value + more


class BoolNot(object):
    """ Represents the boolean operator NOT
    """
    def __init__(self, data):
        self.condition = data[0][1]
        if isinstance(self.condition, Condition) and self.condition.name not in params:
            params.update({self.condition.fullname: self.condition.value})

    def filter(self, DataModelClass):
        """ Return the operator as a SQLAlchemy not_() condition
        """
        if not isinstance(self.condition, FxnCondition):
            return not_(self.condition.filter(DataModelClass))

    def __repr__(self):
        return 'not_(' + repr(self.condition) + ')'


class BoolAnd(object):
    """ Represents the boolean operator AND
    """
    def __init__(self, data):
        self.conditions = []
        for condition in data[0]:
            if condition and condition != 'and':
                if isinstance(condition, FxnCondition):
                    functions.append(condition)
                else:
                    self.conditions.append(condition)
                if isinstance(condition, Condition) and condition.name not in params:
                    params.update({condition.fullname: condition.value})

    def filter(self, DataModelClass):
        """ Return the operator as a SQLAlchemy and_() condition
        """
        conditions = [condition.filter(DataModelClass) for condition in self.conditions if not isinstance(condition, FxnCondition)]
        return and_(*conditions)  # * converts list to argument sequence

    def removeFunctions(self):
        ''' remove the fxn conditions '''
        self.conditions = [condition for condition in self.conditions if not isinstance(condition, FxnCondition)]

    def __repr__(self):
        return 'and_(' + ', '.join([repr(condition) for condition in self.conditions]) + ')'


class BoolOr(object):
    """ Represents the boolean operator OR
    """
    def __init__(self, data):
        self.conditions = []
        for condition in data[0]:
            if condition and condition != 'or':
                if isinstance(condition, FxnCondition):
                    functions.append(condition)
                else:
                    self.conditions.append(condition)
                if isinstance(condition, Condition) and condition.name not in params:
                    params.update({condition.fullname: condition.value})

    def filter(self, DataModelClass):
        """ Return the operator as a SQLAlchemy or_() condition
        """
        conditions = [condition.filter(DataModelClass) for condition in self.conditions if not isinstance(condition, FxnCondition)]
        return or_(*conditions)  # * converts list to argument sequence

    def __repr__(self):
        return 'or_(' + ', '.join([repr(condition) for condition in self.conditions]) + ')'

    def removeFunctions(self):
        ''' remove the fxn conditions '''
        self.conditions = [condition for condition in self.conditions if not isinstance(condition, FxnCondition)]

# ***** Define the boolean condition expressions *****

# Define expression elements
LPAR = pp.Suppress('(')
RPAR = pp.Suppress(')')
number = pp.Regex(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
name = pp.Word(pp.alphas + '._', pp.alphanums + '._')
operator = pp.Regex("==|!=|<=|>=|<|>|=")
value = pp.Word(pp.alphanums + '-_.*') | pp.QuotedString('"') | number

whereexp = pp.Forward()
# condition
condition = pp.Group(name + operator + value)
condition.setParseAction(Condition)
# between condition
between_cond = pp.Group(name + pp.CaselessLiteral('between') + value + pp.CaselessLiteral('and') + value)
between_cond.setParseAction(Condition)
# fxn condition
function_call = pp.Group(pp.Word(pp.alphas) + LPAR + condition + RPAR)
fxn_cond = pp.Group(function_call + operator + value)
fxn_cond.setParseAction(FxnCondition)
# combine
wherecond = condition | fxn_cond | between_cond
whereexp <<= wherecond

# Define the expression as a hierarchy of boolean operators
# with the following precedence: NOT > AND > OR
expression_parser = pp.operatorPrecedence(whereexp, [
    (pp.CaselessLiteral("not"), 1, pp.opAssoc.RIGHT, BoolNot),
    (pp.CaselessLiteral("and"), 2, pp.opAssoc.LEFT, BoolAnd),
    (pp.CaselessLiteral("or"), 2, pp.opAssoc.LEFT, BoolOr),
])

params = {}
uniqueparams = []
functions = []


def parse_boolean_search(boolean_search):
    """ Parses the boolean search expression into a hierarchy of boolean operators.
        Returns a BoolNot or BoolAnd or BoolOr object.
    """
    global params, functions, uniqueparams
    params = {}
    uniqueparams = []
    functions = []
    try:
        expression = expression_parser.parseString(boolean_search)[0]
    except ParseException as e:
        raise BooleanSearchException("Syntax error at offset %(offset)s." % dict(offset=e.col))
    else:
        expression.params = params
        expression.uniqueparams = list(set(uniqueparams))
        expression.functions = functions
        return expression





#!/usr/bin/env python
# encoding: utf-8
#
# rss.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     11 Apr 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function

from flask import jsonify
from flask_classy import route

from marvin.tools.rss import RSS
from marvin.api.base import BaseView, arg_validate as av
from marvin.core.exceptions import MarvinError
from marvin.utils.general import parseIdentifier


def _getRSS(name, **kwargs):
    """Retrieves a RSS Marvin object."""

    rss = None
    results = {}

    # Pop the release to remove a duplicate input to Maps
    release = kwargs.pop('release', None)

    # parse name into either mangaid or plateifu
    try:
        idtype = parseIdentifier(name)
    except Exception as e:
        results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(e))
        return rss, results

    try:
        if idtype == 'plateifu':
            plateifu = name
            mangaid = None
        elif idtype == 'mangaid':
            mangaid = name
            plateifu = None
        else:
            raise MarvinError('invalid plateifu or mangaid: {0}'.format(idtype))

        rss = RSS(mangaid=mangaid, plateifu=plateifu, mode='local', release=release)
        results['status'] = 1
    except Exception as e:
        results['error'] = 'Failed to retrieve RSS {0}: {1}'.format(name, str(e))

    return rss, results


class RSSView(BaseView):
    """Class describing API calls related to RSS files."""

    route_base = '/rss/'

    @route('/<name>/', methods=['GET', 'POST'], endpoint='getRSS')
    @av.check_args()
    def get(self, args, name):
        """This method performs a get request at the url route /rss/<id>.

        .. :quickref: RSS; Get an RSS given a plate-ifu or mangaid

        :param name: The name of the cube as plate-ifu or mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string empty: the data dict is empty
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/rss/8485-1901/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {}
           }

        """

        # Pop any args we don't want going into Rss
        args = self._pop_args(args, arglist='name')

        rss, results = _getRSS(name, **args)
        self.update_results(results)

        if rss:
            # For now we don't return anything here, maybe later.
            self.results['data'] = {}

        return jsonify(self.results)

    @route('/<name>/fibers/', methods=['GET', 'POST'], endpoint='getRSSAllFibers')
    @av.check_args()
    def getAllFibers(self, args, name):
        """Returns a list of all the flux, ivar, mask, and wavelength arrays for all fibres.

        .. :quickref: RSS; Get a list of flux, ivar, mask, and wavelength arrays for all fibers

        :param name: The name of the cube as plate-ifu or mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json list rssfiber: the flux, ivar, mask arrays for the given rssfiber index
        :json list wavelength: the wavelength arrays for all fibers
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/rss/8485-1901/fibers/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"wavelength": [3621.6, 3622.43, 3623.26, ...],
                       "0": [flux, ivar, mask],
                       "1": [flux, ivar, mask],
                       ...
                       "170": [flux, ivar, mask]
              }
           }

        """

        # Pop any args we don't want going into Rss
        args = self._pop_args(args, arglist='name')

        rss, results = _getRSS(name, **args)
        self.update_results(results)

        if rss:
            self.results['data'] = {}
            self.results['data']['wavelength'] = rss[0].wavelength.tolist()

            for ii, fiber in enumerate(rss):
                flux = fiber.flux.tolist()
                ivar = fiber.ivar.tolist()
                mask = fiber.mask.tolist()
                self.results['data'][ii] = [flux, ivar, mask]

        return jsonify(self.results)



#!/usr/bin/env python
# encoding: utf-8

import numpy as np

from flask_classy import route
from flask import jsonify, request, Response
import json

from marvin.api.base import BaseView, arg_validate as av
from marvin.core.exceptions import MarvinError
from marvin.utils.general import parseIdentifier
from marvin.tools.cube import Cube

from brain.utils.general import parseRoutePath
from brain.core.exceptions import BrainError

''' stuff that runs server-side '''


def _getCube(name, **kwargs):
    ''' Retrieve a cube using marvin tools '''

    # Pop the release to remove a duplicate input to Maps
    release = kwargs.pop('release', None)

    cube = None
    results = {}

    # parse name into either mangaid or plateifu
    try:
        idtype = parseIdentifier(name)
    except Exception as ee:
        results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(ee))
        return cube, results

    try:
        if idtype == 'plateifu':
            plateifu = name
            mangaid = None
        elif idtype == 'mangaid':
            mangaid = name
            plateifu = None
        else:
            raise MarvinError('invalid plateifu or mangaid: {0}'.format(idtype))

        cube = Cube(mangaid=mangaid, plateifu=plateifu, mode='local', release=release)
        results['status'] = 1
    except Exception as ee:
        results['error'] = 'Failed to retrieve cube {0}: {1}'.format(name, str(ee))

    return cube, results


class CubeView(BaseView):
    ''' Class describing API calls related to MaNGA Cubes '''

    route_base = '/cubes/'
    # decorators = [av.check_args()]

    def index(self):
        '''Returns general cube info

        .. :quickref: Cube; Get general cube info

        :query string release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: data message
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/cubes/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": "this is a cube!"
           }

        '''
        args = av.manual_parse(self, request)
        self.results['status'] = 1
        self.results['data'] = 'this is a cube!'
        return jsonify(self.results)

    @route('/<name>/', methods=['GET', 'POST'], endpoint='getCube')
    @av.check_args()
    def get(self, args, name):
        '''Returns the necessary information to instantiate a cube for a given plateifu.

        .. :quickref: Cube; Get a cube given a plate-ifu or mangaid

        :param name: The name of the cube as plate-ifu or mangaid
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string plateifu: id of cube
        :json string mangaid: mangaid of cube
        :json float ra: RA of cube
        :json float dec: Dec of cube
        :json string header: the cube header as a string
        :json float redshift: the cube redshift
        :json list shape: the cube shape [x, y]
        :json list wavelength: the cube wavelength array
        :json string wcs_header: the cube wcs_header as a string
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/cubes/8485-1901/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateifu": "8485-1901",
                    "mangaid": "1-209232",
                    "ra": 232.544703894,
                    "dec": 48.6902009334,
                    "header": "XTENSION= 'IMAGE', NAXIS=3, .... END",
                    "wcs_header": "WCSAXES = 3 / Number of coordindate axes .... END",
                    "redshift": 0.0407447,
                    "shape": [34, 34],
                    "wavelength": [3621.6, 3622.43,...,10353.8]
              }
           }

        '''

        # Pop any args we don't want going into Cube
        args = self._pop_args(args, arglist='name')

        cube, res = _getCube(name, **args)
        self.update_results(res)

        if cube:

            try:
                nsa_data = cube.nsa
            except (MarvinError, BrainError):
                nsa_data = None

            wavelength = (cube.wavelength.tolist() if isinstance(cube.wavelength, np.ndarray)
                          else cube.wavelength)
            self.results['data'] = {'plateifu': name,
                                    'mangaid': cube.mangaid,
                                    'ra': cube.ra,
                                    'dec': cube.dec,
                                    'header': cube.header.tostring(),
                                    'redshift': nsa_data.z if nsa_data else -9999,
                                    'shape': cube.shape,
                                    'wavelength': wavelength,
                                    'wcs_header': cube.wcs.to_header_string()}

        return jsonify(self.results)

    @route('/<name>/extensions/<cube_extension>/', methods=['GET', 'POST'],
           endpoint='getCubeExtension')
    @av.check_args()
    def getCubeExtension(self, args, name, cube_extension):
        ''' Returns the 3d flux, ivar, or mask for a cube given a plateifu/mangaid.

        .. :quickref: Cube; Get the full cube 3d flux, ivar, or mask given a plate-ifu or mangaid

        :param name: The name of the cube as plate-ifu or mangaid
        :param cube_extension: The name of the cube extension.  Either flux, ivar, or mask.
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string cube_extension: the 3d data for the specified extension
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin2/api/cubes/8485-1901/extensions/flux/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"cube_extension": [[0,0,..0], [], ... [0, 0, 0,... 0]]
              }
           }
        '''

        # Pass the args in and get the cube
        args = self._pop_args(args, arglist='name')
        cube, res = _getCube(name, **args)
        self.update_results(res)

        if cube:
            self.results['data'] = {'cube_extension':
                                    cube._getExtensionData(cube_extension.upper()).tolist()}

        return Response(json.dumps(self.results), mimetype='application/json')



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-07 16:40:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-06 16:45:23

from __future__ import print_function, division, absolute_import
from marvin.tests.api.conftest import ApiPage
from marvin.tools.query import Query
from marvin.tools.query.query_utils import bestparams
import pytest
import yaml
import os

# @pytest.fixture()
# def page(client, request, init_api):
#     blue, endpoint = request.param
#     page = ApiPage(client, 'api', endpoint)
#     yield page

query_data = yaml.load(open(os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/query_test_data.dat'))))


@pytest.fixture()
def params(release):
    return {'release': release}


@pytest.fixture()
def data(release):
    return query_data[release]


def get_query_params(release, paramdisplay):
    q = Query(mode='local', release=release)
    if paramdisplay == 'best':
        qparams = bestparams
    else:
        qparams = q.get_available_params('all')
    return qparams


@pytest.mark.parametrize('page', [('api', 'QueryView:index')], ids=['query'], indirect=True)
class TestQueryView(object):

    def test_get_query_success(self, page, params):
        page.load_page('get', page.url, params=params)
        data = 'this is a query'
        page.assert_success(data)


@pytest.mark.parametrize('page', [('api', 'querycubes')], ids=['querycubes'], indirect=True)
class TestQueryCubes(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('searchfilter', [('nsa.z < 0.1')])
    def test_query_success(self, page, params, reqtype, searchfilter, data):
        params.update({'searchfilter': searchfilter})
        expdata = data['queries'][searchfilter]['top5']
        page.load_page(reqtype, page.url, params=params)
        page.assert_success(expdata, issubset=True)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                       (None, 'searchfilter', 'Missing data for required field.')],
                             ids=['norelease', 'nosearchfilter'])
    def test_query_failure(self, page, reqtype, params, name, missing, errmsg):
        page.route_no_valid_params(page.url, missing, reqtype=reqtype, errmsg=errmsg)


@pytest.mark.parametrize('page', [('api', 'getparams')], ids=['getparams'], indirect=True)
class TestQueryGetParams(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('paramdisplay', [('all'), ('best')])
    def test_getparams_success(self, release, page, params, reqtype, paramdisplay):
        params.update({'paramdisplay': paramdisplay})
        expdata = get_query_params(release, paramdisplay)
        page.load_page(reqtype, page.url, params=params)
        page.assert_success(expdata, keys=True)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg', [(None, 'release', 'Missing data for required field.'),
                                                       (None, 'paramdisplay', 'Missing data for required field.')],
                             ids=['norelease', 'noparamdisplay'])
    def test_plateifu_failure(self, page, reqtype, params, name, missing, errmsg):
        page.route_no_valid_params(page.url, missing, reqtype=reqtype, errmsg=errmsg)





# License information goes here
# -*- coding: utf-8 -*-
#
# $Id: system_call.py 157318 2014-06-29 12:10:03Z weaver $
#
"""
"""
#
# This will help with 2to3 support.
#
from __future__ import absolute_import, division, print_function, unicode_literals
#
#
#
def system_call(command,test=False,logger=None,logall=False,message=None,outname=None,errname=None):
    """A convenient wrapper to log and perform system calls.

    Parameters
    ----------
    command : str
        The system command to run.  It will be split internally by shlex.split().
    test : bool, optional
        Set this to true to not actually run the commands.
    logger : logging.logging, optional
        If passed a logging object, diagnostic output will be set to this object.
    logall : bool, optional
        Set this to true to always log stdout and stderr (at level DEBUG).
        Otherwise stdout and stderr will only be logged for nonzero status
    message : str, optional
        Call logger.critical(message) with this message and then call
        sys.exit(status).
    outname : str, optional
        If set, use ``outname`` as the name of a file to contain stdout.  It
        is the responsibility of the caller of this function to clean up
        the resulting file.  Otherwise a temporary file will be used.
    errname : str, optional
        Same as ``outname``, but will contain stderr.

    Returns
    -------
    (status,out,err) : tuple
        The exit status, stdout and stderr of the process

    Examples
    --------
    >>> status,out,err = transfer.common.system_call('date')
    """
    from math import log10
    from sys import exit
    from subprocess import Popen
    from shlex import split
    from tempfile import TemporaryFile
    from time import time, sleep
    if logger is not None:
        logger.debug(command)
    status = 0
    out = ''
    err = ''
    if not test:
        if outname is None:
            outfile = TemporaryFile()
        else:
            outfile = open(outname,'w+')
        if errname is None:
            errfile = TemporaryFile()
        else:
            errfile = open(errname,'w+')
        proc = Popen(split(str(command)),stdout=outfile,stderr=errfile)
        tstart = time()
        while proc.poll() is None:
            elapsed = time() - tstart
            if elapsed > 500000:
                message = "Process still running after more than 5 days!"
                proc.kill()
                break
            tsleep = 10**(int(log10(elapsed))-1)
            if tsleep < 1:
                tsleep = 1
            sleep(tsleep)
        # proc.wait()
        status = proc.returncode
        outfile.seek(0)
        out = outfile.read()
        errfile.seek(0)
        err = errfile.read()
        outfile.close()
        errfile.close()
        if logger is not None:
            if status == 0 and logall:
                if len(out) > 0:
                    logger.debug('STDOUT = \n' + out)
                if len(err) > 0:
                    logger.debug('STDERR = \n' + err)
            if status != 0:
                logger.error('status = {0}'.format(status))
                if len(out) > 0:
                    logger.error('STDOUT = \n' + out)
                if len(err) > 0:
                    logger.error('STDERR = \n' + err)
                if message is not None:
                    logger.critical(message)
                    exit(status)
    return (status,out,err)



#!/usr/bin/env python

import os
import re

import pytest
import numpy as np
from astropy.io import fits
from astropy import wcs

from marvin import config, marvindb
from marvin.tools.cube import Cube
from marvin.core.core import DotableCaseInsensitive
from marvin.core.exceptions import MarvinError
from marvin.tests import skipIfNoDB, marvin_test_if


@pytest.fixture(autouse=True)
def skipbins(galaxy):
    if galaxy.bintype not in ['SPX', 'NONE']:
        pytest.skip('Skipping all bins for Cube tests')
    if galaxy.template not in ['MILES-THIN', 'GAU-MILESHC']:
        pytest.skip('Skipping all templates for Cube tests')


class TestCube(object):

    def test_cube_loadfail(self):
        with pytest.raises(AssertionError) as cm:
            Cube()
        assert 'Enter filename, plateifu, or mangaid!' in str(cm.value)

    def test_cube_load_from_local_file_by_filename_success(self, galaxy):
        cube = Cube(filename=galaxy.cubepath)
        assert cube is not None
        assert os.path.realpath(galaxy.cubepath) == cube.filename

    def test_cube_load_from_local_file_by_filename_fail(self):
        with pytest.raises(MarvinError):
            Cube(filename='not_a_filename.fits')

    def test_cube_load_from_local_database_success(self, galaxy):
        """Tests for Cube Load by Database."""
        cube = Cube(mangaid=galaxy.mangaid)
        assert cube is not None
        assert galaxy.mangaid == cube.mangaid
        assert galaxy.plate == cube.plate
        assert galaxy.dec == cube.dec
        assert galaxy.ra == cube.ra

    @pytest.mark.parametrize('plateifu, mode, errmsg',
                             [('8485-0923', 'local', 'Could not retrieve cube for plate-ifu 8485-0923: No Results Found')],
                             ids=['noresults'])
    def test_cube_from_db_fail(self, plateifu, mode, errmsg):
        with pytest.raises(MarvinError) as cm:
            c = Cube(plateifu=plateifu, mode=mode)
        assert errmsg in str(cm.value)

    # @pytest.mark.slow
    @marvin_test_if(mark='include', cube={'plateifu': '8485-1901'})
    def test_cube_flux(self, cube):
        assert cube.flux is not None
        assert isinstance(cube.flux, np.ndarray)

    @pytest.mark.parametrize('monkeyconfig',
                             [('release', 'MPL-5')],
                             ids=['mpl5'], indirect=True)
    def test_cube_remote_drpver_differ_from_global(self, galaxy, monkeyconfig):

        assert config.release == 'MPL-5'
        cube = Cube(plateifu=galaxy.plateifu, mode='remote', release='MPL-4')
        assert cube._drpver == 'v1_5_1'
        assert cube.header['VERSDRP3'].strip() == 'v1_5_0'

    def test_cube_redshift(self, cube, galaxy):
        assert cube.data_origin == cube.exporigin
        redshift = cube.nsa.redshift if cube.release == 'MPL-4' and cube.data_origin == 'file' else cube.nsa.z
        assert pytest.approx(redshift, galaxy.redshift)

    def test_release(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        assert cube.release == galaxy.release

    def test_set_release_fails(self, galaxy):
        cube = Cube(plateifu=galaxy.plateifu)
        with pytest.raises(MarvinError) as ee:
            cube.release = 'a'
        assert 'the release cannot be changed' in str(ee.value)

    def test_load_filename_does_not_exist(self):
        """Tries to load a file that does not exist, in auto mode."""
        with pytest.raises(MarvinError) as ee:
            Cube(filename='hola.fits', mode='auto')

        assert re.match(r'input file .+hola.fits not found', str(ee.value)) is not None

    def test_load_filename_remote(self):
        """Tries to load a filename in remote mode and fails."""
        with pytest.raises(MarvinError) as ee:
            Cube(filename='hola.fits', mode='remote')

        assert 'filename not allowed in remote mode' in str(ee.value)


class TestGetSpaxel(object):

    def _dropNones(self, **kwargs):
        for k, v in list(kwargs.items()):
            if v is None:
                del kwargs[k]
        return kwargs

    @pytest.mark.parametrize('x, y, ra, dec, excType, message',
        [(1, None, 1, None, AssertionError, 'Either use (x, y) or (ra, dec)'),
         (1, None, 1, 1, AssertionError, 'Either use (x, y) or (ra, dec)'),
         (1, None, None, None, AssertionError, 'Specify both x and y'),
         (None, 1, None, None, AssertionError, 'Specify both x and y'),
         (None, None, 1, None, AssertionError, 'Specify both ra and dec'),
         (None, None, None, 1, AssertionError, 'Specify both ra and dec'),
         (None, None, None, None, ValueError, 'You need to specify either (x, y) or (ra, dec)'),
         (-50, 1, None, None, MarvinError, 'some indices are out of limits'),
         (50, 1, None, None, MarvinError, 'some indices are out of limits'),
         (1, -50, None, None, MarvinError, 'some indices are out of limits'),
         (1, 50, None, None, MarvinError, 'some indices are out of limits'),
         (None, None, 1., 1., MarvinError, 'some indices are out of limits'),
         (None, None, 100, 60, MarvinError, 'some indices are out of limits'),
         (None, None, 232.546383, 1., MarvinError, 'some indices are out of limits'),
         (None, None, 1., 48.6883954, MarvinError, 'some indices are out of limits')],
        ids=['x-ra', 'x-ra-dec', 'x', 'y', 'ra', 'dec', 'no-inputs', '-50-1', '50-1', '1--50',
             '1-50', '1-1', '100-60', '232.5-1', '1-48.6'])
    def test_getSpaxel_inputs(self, galaxy, x, y, ra, dec, excType, message):
        """Tests exceptions when getSpaxel gets inappropriate inputs."""
        kwargs = self._dropNones(x=x, y=y, ra=ra, dec=dec)

        with pytest.raises(excType) as ee:
            cube = Cube(plateifu=galaxy.plateifu, release=galaxy.release)
            cube.getSpaxel(**kwargs)

        assert message in str(ee.value)

    @pytest.mark.parametrize('coord, xyorig',
                             [('xy', 'lower'),
                              ('xy', 'center'),
                              ('radec', None)])
    def test_getSpaxel_flux(self, cube, galaxy, coord, xyorig):
        if coord == 'xy':
            x = galaxy.spaxel['x'] if xyorig == 'lower' else galaxy.spaxel['x_cen']
            y = galaxy.spaxel['y'] if xyorig == 'lower' else galaxy.spaxel['y_cen']
            params = {'x': x, 'y': y, 'xyorig': xyorig}
        elif coord == 'radec':
            ra = galaxy.spaxel['ra']
            dec = galaxy.spaxel['dec']
            params = {'ra': ra, 'dec': dec}

        spaxel = cube.getSpaxel(**params)
        flux = spaxel.spectrum.flux
        assert pytest.approx(flux[galaxy.spaxel['specidx']], galaxy.spaxel['flux'])

    @pytest.mark.parametrize('monkeyconfig',
                             [('sasurl', 'http://www.averywrongurl.com')],
                             ids=['wrongurl'], indirect=True)
    def test_getSpaxel_remote_fail_badresponse(self, monkeyconfig):
        assert config.urlmap is not None

        with pytest.raises(MarvinError) as cm:
            Cube(mangaid='1-209232', mode='remote')

        assert 'Failed to establish a new connection' in str(cm.value)

    def test_getSpaxel_remote_drpver_differ_from_global(self, galaxy):
        config.setMPL('MPL-5')
        assert config.release == 'MPL-5'

        cube = Cube(plateifu=galaxy.plateifu, mode='remote', release='MPL-4')
        expected = galaxy.spaxel['flux']

        spectrum = cube.getSpaxel(ra=galaxy.spaxel['ra'], dec=galaxy.spaxel['dec']).spectrum
        assert pytest.approx(spectrum.flux[galaxy.spaxel['specidx']], expected)

    def test_getspaxel_matches_file_db_remote(self, galaxy):

        cube_file = Cube(filename=galaxy.cubepath)
        cube_db = Cube(plateifu=galaxy.plateifu)
        cube_api = Cube(plateifu=galaxy.plateifu, mode='remote')

        assert cube_file.data_origin == 'file'
        assert cube_db.data_origin == 'db'
        assert cube_api.data_origin == 'api'

        xx = galaxy.spaxel['x']
        yy = galaxy.spaxel['y']
        spec_idx = galaxy.spaxel['specidx']
        flux = galaxy.spaxel['flux']
        ivar = galaxy.spaxel['ivar']
        mask = galaxy.spaxel['mask']

        spaxel_slice_file = cube_file[yy, xx]
        spaxel_slice_db = cube_db[yy, xx]
        spaxel_slice_api = cube_api[yy, xx]

        assert pytest.approx(spaxel_slice_file.spectrum.flux[spec_idx], flux)
        assert pytest.approx(spaxel_slice_db.spectrum.flux[spec_idx], flux)
        assert pytest.approx(spaxel_slice_api.spectrum.flux[spec_idx], flux)

        assert pytest.approx(spaxel_slice_file.spectrum.ivar[spec_idx], ivar)
        assert pytest.approx(spaxel_slice_db.spectrum.ivar[spec_idx], ivar)
        assert pytest.approx(spaxel_slice_api.spectrum.ivar[spec_idx], ivar)

        assert pytest.approx(spaxel_slice_file.spectrum.mask[spec_idx], mask)
        assert pytest.approx(spaxel_slice_db.spectrum.mask[spec_idx], mask)
        assert pytest.approx(spaxel_slice_api.spectrum.mask[spec_idx], mask)

        xx_cen = galaxy.spaxel['x_cen']
        yy_cen = galaxy.spaxel['y_cen']

        spaxel_getspaxel_file = cube_file.getSpaxel(x=xx_cen, y=yy_cen)
        spaxel_getspaxel_db = cube_db.getSpaxel(x=xx_cen, y=yy_cen)
        spaxel_getspaxel_api = cube_api.getSpaxel(x=xx_cen, y=yy_cen)

        assert pytest.approx(spaxel_getspaxel_file.spectrum.flux[spec_idx], flux)
        assert pytest.approx(spaxel_getspaxel_db.spectrum.flux[spec_idx], flux)
        assert pytest.approx(spaxel_getspaxel_api.spectrum.flux[spec_idx], flux)

        assert pytest.approx(spaxel_getspaxel_file.spectrum.ivar[spec_idx], ivar)
        assert pytest.approx(spaxel_getspaxel_db.spectrum.ivar[spec_idx], ivar)
        assert pytest.approx(spaxel_getspaxel_api.spectrum.ivar[spec_idx], ivar)

        assert pytest.approx(spaxel_getspaxel_file.spectrum.mask[spec_idx], mask)
        assert pytest.approx(spaxel_getspaxel_db.spectrum.mask[spec_idx], mask)
        assert pytest.approx(spaxel_getspaxel_api.spectrum.mask[spec_idx], mask)


class TestWCS(object):

    def test_wcs(self, cube):
        assert cube.data_origin == cube.exporigin
        assert isinstance(cube.wcs, wcs.WCS)
        comp = cube.wcs.wcs.pc if cube.data_origin == 'api' else cube.wcs.wcs.cd
        assert pytest.approx(comp[1, 1], 0.000138889)


class TestPickling(object):

    def test_pickling_file(self, temp_scratch, galaxy):
        cube = Cube(filename=galaxy.cubepath)
        assert cube.data_origin == 'file'
        assert isinstance(cube, Cube)
        assert cube.data is not None

        assert not os.path.isfile(galaxy.cubepath[0:-7] + 'mpf')
        cube_file = temp_scratch.join('test_cube.mpf')
        path = cube.save(str(cube_file))
        assert cube_file.check() is True
        assert cube.data is not None

        cube = None
        assert cube is None

        cube_restored = Cube.restore(str(cube_file))
        assert cube_restored.data_origin == 'file'
        assert isinstance(cube_restored, Cube)
        assert cube_restored.data is not None

    def test_pickling_file_custom_path(self, temp_scratch, galaxy):
        cube = Cube(filename=galaxy.cubepath)
        assert cube.data_origin == 'file'
        assert isinstance(cube, Cube)
        assert cube.data is not None

        test_path = temp_scratch.join('cubepickle').join('test_cube.mpf')
        assert test_path.check(file=1) is False

        path = cube.save(path=str(test_path))
        assert test_path.check(file=1) is True
        assert path == os.path.realpath(os.path.expanduser(str(test_path)))

        cube_restored = Cube.restore(str(test_path), delete=True)
        assert cube_restored.data_origin == 'file'
        assert isinstance(cube_restored, Cube)
        assert cube_restored.data is not None

        assert not os.path.exists(path)

    def test_pickling_db(self, galaxy, temp_scratch):
        cube = Cube(plateifu=galaxy.plateifu)
        assert cube.data_origin == 'db'

        file = temp_scratch.join('test_cube_db.mpf')
        with pytest.raises(MarvinError) as cm:
            cube.save(str(file))

        assert 'objects with data_origin=\'db\' cannot be saved.' in str(cm.value)

    def test_pickling_api(self, temp_scratch, galaxy):
        cube = Cube(plateifu=galaxy.plateifu, mode='remote')
        assert cube.data_origin == 'api'
        assert isinstance(cube, Cube)
        assert cube.data is None

        test_path = temp_scratch.join('test_cube_api.mpf')

        path = cube.save(str(test_path))
        assert test_path.check() is True

        cube = None
        assert cube is None

        cube_restored = Cube.restore(str(test_path))
        assert cube_restored.data_origin == 'api'
        assert isinstance(cube_restored, Cube)
        assert cube_restored.data is None



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-03-24 12:22:07
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-04-04 10:30:54

from __future__ import print_function, division, absolute_import


class TestRsyncAdd(object):

    def test_rsync_add_link(self, rsync_add):
        task = rsync_add.initial_stream.task[0]
        assert rsync_add.location in task['location']


class TestRsyncStream(object):

    def test_rsync_setstream_task(self, rsync_set):
        tasks = rsync_set.stream.task
        task = tasks[0]
        assert rsync_set.location in task['location']
        assert len(tasks) == rsync_set.count



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-04-06 16:41:53
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-04-24 22:15:14

from __future__ import print_function, division, absolute_import
import time
import os
import pytest
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from marvin.tests.web.frontend.models import IndexPage, SearchPage


@pytest.fixture()
def page(driver, base_url):
    page = IndexPage(driver, root_uri=base_url)
    return page


@pytest.mark.xfail()
@pytest.mark.timeout(60)
@pytest.mark.usefixtures('live_server')
class TestIndexPage(object):
    ''' Tests for the main Index page '''

    def test_title(self, page):
        assert 'Marvin' in page.w.title

    def test_goto_random(self, page):
        assert 'Marvin' in page.w.title
        page.imagepage.click()
        time.sleep(1)
        assert 'random' in page.w.current_url

    def test_goto_search(self, page):
        page.searchpage.click()
        time.sleep(1)
        assert 'Search' in page.w.title
        assert 'search' in page.w.current_url
        results = page.w.find_elements(By.ID, "search_results")
        assert len(results) == 0


@pytest.fixture()
def search_page(driver, base_url):
    page = SearchPage(driver, root_uri=base_url)
    page.get('search/')
    return page


@pytest.mark.xfail()
@pytest.mark.timeout(60)
@pytest.mark.usefixtures('live_server')
class TestSearchPage(object):
    ''' Tests for the main Search page '''

    def test_title(self, search_page):
        assert 'Search' in search_page.w.title

    def test_search(self, search_page):
        assert search_page.results is None
        search_page.searchbox = 'nsa.z < 0.1'
        search_page.searchbox = Keys.RETURN
        assert search_page.results is not None
        assert search_page.table is not None
